"""
Generalized-LR parsing amounts to direct simulation of a non-deterministic handle-finding
automaton (HFA). There are two key design decisions: How to represent the stack(s), and how
to orchestrate the operation of semantic actions. In other tools there are various approaches
taken. I'd like to survey those approaches, comment on their strengths and weaknesses, and
provide the ability to choose among them according to your specific parsing needs.

It's about this point that the necessary interface to a non-deterministic parse table is
made plain: GOTO can work just like in the deterministic case (because it's not affected
by non-determinism) but ACTION needs the additional skill to store non-deterministic
entries. There are a thousand ways to do it. One way? Any instruction greater than
the maximum valid state number refers instead to a "non-determinism" entry consisting
of a list of shift/reduce instructions.

A non-deterministic stack can be understood as a directed acyclic graph with a single "sink"
node representing the bottom of the stack. Each node is labeled with an HFA state and some
analogue to a "semantic value", the details of which depend on how clever you're being.
The reaching-symbols along each path from a "source" node to the sink represents a potential
viable prefix of a rightmost derivation. Each subsequent token may invalidate some paths and
create additional ambiguity along others. If all paths die, the input is not a member of the
language. If more than one path remains when the input and end-marker are all consumed, then
the input has more than one valid interpretation. Depending on your application, such a
situation may or may not be a problem.

At relatively low levels of ambiguity, it's reasonable to let each state have one distinct
semantic value, computed bottom-up during the parse: the graph is thus an inverted tree.
This approach is easy to understand, easy to code, and potentially leads to exponential
behavior. Still, it's usually just fine for grammars that are just barely out of reach for
a deterministic parse table. In this strategy, either the semantic actions must be pure
functions or else the parser must perform certain gymnastics to delay the invocation of
actions until any ambiguity is resolved.

At the opposite extreme, the top of the stack is organized to contain at most one node per
state-id, and all intermediate semantic values are structured to reflect ambiguity. The result
of such a parse represents a non-deterministic tree whose yield is the sequence of terminals.
This keeps algorithmic complexity down during parsing, which is great for highly-ambiguous
grammars, but then the semantic analysis code must deal sensibly with that ambiguity.

An interesting middle ground says the top-of-stack can be ambiguous a'la case two (the previous
paragraph) but the non-determinism must be resolved for each node bottom-up so that a given
node has at most one semantic value.
"""

from ..support import interfaces

class AbstractGeneralizedParser:
	"""
	Before I get too deep into it, let's lay out the general structure of a generalized parse:
	"""
	def __init__(self, table: interfaces.ParseTable, combine, language=None):
		""" Please note this takes a driver not a combiner: it does its own selection of arguments from the stack. """
		self._table = table
		self._combine = combine
		self._nr_states = table.get_split_offset()
		self.reset(table.get_initial(language))
	
	def reset(self, initial_state):
		""" Configure the initial stack situation for the given initial automaton state. """
		raise NotImplementedError(type(self))
	
	def consume(self, terminal, semantic):
		""" Call this from your scanning loop. """
		raise NotImplementedError(type(self))

	def finish(self) -> list:
		"""
		Call this after the last token to wrap up and
		:return: a valid semantic value for the parse.
		"""
		raise NotImplementedError(type(self))


class BruteForceAndIgnorance(AbstractGeneralizedParser):
	"""
	There is an old adage in software development: when in doubt, use brute force. Accordingly,
	the first implementation will be simple, easy to code, slow, and vulnerable to exponential
	behavior. However, it will also provide a fine basis of comparison for other approaches.
	
	This is the "inverse tree" approach: each node is a state, a predecessor, and a semantic
	value for the reaching-symbol -- except for the sink, which is special. The top-of-stack
	is just a list of currently-viable sub-stacks. This is usually sufficient for cases that
	are actually unambiguous but just not quite LR(1)-deterministic.
	
	Big problems with this approach, in no particular order, include:
		1. It fails to coalesce parse stacks that reduce to a common state.
		2. It eagerly calls reductions, even for dead-end parse attempts.
	
	So you can read what's going on:
		self.__tos contains nodes representing currently active parses.
		self.__next gets filled with the results of SHIFT actions.
		Each node is a tuple of (state_id, prior_node, semantic_value), accessed
			by field numbers given as the constants below.
	"""
	
	NODE_STATE = 0
	NODE_PRIOR = 1
	NODE_SEMANTIC = 2
	
	def reset(self, initial_state):
		self.__tos = [(initial_state, None, None)]
	
	def consume(self, terminal, semantic):
		self.__consume(self._table.get_translation(terminal), semantic)
		if not self.__tos: raise interfaces.GeneralizedParseError("Parser died midway at something ungrammatical.")

	def __consume(self, terminal_id, semantic):
		self.__next = []
		while self.__tos:
			top = self.__tos.pop()
			self.__act(self._table.get_action(top[self.NODE_STATE], terminal_id), top, semantic)
		self.__tos = self.__next

	def finish(self) -> list:
		"""
		Call this after the last token to wrap up and
		:return: a valid semantic value for the parse.
		"""
		self.__consume(0, None)
		if self.__tos: return [top[self.NODE_PRIOR][self.NODE_SEMANTIC] for top in self.__tos]
		else: raise interfaces.GeneralizedParseError("Parser recognized a viable prefix, but not a complete sentence.")
	
	def __act(self, action, top, semantic):
		""" There are four kinds of action: die, shift, reduce-and-shift, or split into parallel alternatives. """
		if action == 0: return # This branch of the stack dies.
		elif action < 0: self.__tos.append(self.__reduction(-1 - action, top))
		elif action < self._nr_states: self.__shift(action, top, semantic)
		else:
			for alternative in self._table.get_split(action - self._nr_states):
				self.__act(alternative, top, semantic)
	
	def __shift(self, state_id, top, semantic):
		shift = state_id, top, semantic
		while True:
			action = self._table.interactive_step(shift[self.NODE_STATE])
			if action < 0: shift = self.__reduction(-1 - action, shift)
			else: break
		self.__next.append(shift)
	
	def __reduction(self, rule_id, top):
		nonterminal_id, length, cid, view = self._table.get_rule(rule_id)
		if cid < 0: semantic = self.__view(top, (cid,))[0]
		else:
			args = self.__view(top, view)
			semantic = self._combine(cid, args)
		while length > 0:
			length -= 1
			top = top[self.NODE_PRIOR]
		return self._table.get_goto(top[self.NODE_STATE], nonterminal_id), top, semantic
	
	def __view(self, top, view):
		"""
		Recall that each element of view is a negative offset from the end of a notional
		linked-list-style stack, so in particular -1 is top-of-stack, and also these are
		presently constrained to appear in increasing order (starting negative and
		growing closer to zero).
		"""
		result = []
		depth = -1
		for seeking in reversed(view):
			while depth > seeking:
				depth -= 1
				top = top[self.NODE_PRIOR]
			result.append(top[self.NODE_SEMANTIC])
		result.reverse()
		return result

def gss_trial_parse(table: interfaces.ParseTable, sentence, *, language=None):
	"""
	By now we've seen two ways each to do both recognition and semantic recovery:
	
	In module parsing.shift_reduce, the trial_parse(...) and parse(...) functions
	show the linear-time approach appropriate to unambiguous parse tables -- i.e.
	those which either have no CF-grammar inadequacies, or for which all such are
	suitably resolved with operator-precedence/associativity declarations.
	
	In module parsing.automata, the trial_parse(...) method of class HFA shows
	a simple recognizer, and the above-styled "brute force and ignorance" class
	shows a simple semantic recovery method, both suitable principally for getting
	our feet wet with the issues involved in dealing with ambiguous parse tables.
	
	It's time to do better. In 1988, Tomita showed the world how to use a so-called
	"graph-structured stack" to avoid the exponential behaviour associated with
	the brute-force approach. I'd like to implement that approach -- first with
	a recognizer and only then, once the essential algorithm is absolutely clear,
	as a full-on parser.
	
	This function is that recognizer.
	
	Interestingly, timings show the BF&I class to be slightly faster than this
	for inputs with only mild levels of ambiguity.
	"""
	nr_states = table.get_split_offset()
	
	class Node: # Nodes do have SOME structure. This should be adequate.
		__slots__ = ["state_id", "edges"]
		def __init__(self, state_id: int, edges:set):
			self.state_id = state_id  # encodes also the reaching symbol
			self.edges = edges
		def all_paths(self, length:int):
			if length < 1: yield self
			else:
				for p in self.edges: yield from p.all_paths(length - 1)
		def __repr__(self): return "<%d / %s>"%(self.state_id, ",".join(str(e.state_id) for e in self.edges))
	
	def act_on(node:Node, instruction:int):
		if instruction == 0: return
		elif instruction < 0: perform_reduction(node, -1-instruction)
		elif instruction < nr_states: # Shift the token:
			# Here, combining happens naturally because `shifts` is keyed to the destination state.
			if instruction in shifts: shifts[instruction].edges.add(node)
			else: shifts[instruction] = Node(instruction, {node})
		
		else: # There are multiple possibilities: split into multiple states as needed.
			for alternative in table.get_split(instruction - nr_states):
				act_on(node, alternative)
	
	def perform_reduction(node:Node, rule_id:int):
		"""
		This is a bit of a mind-bender, because it needs to do the right thing about
		both combining AND local-ambiguity packing. So here's the key insight:
		Local-ambiguity packing closely related to combining, but with respect to
		predecessors rather than the top-of-stack.
		
		Observation:
		The problem gets considerably harder if we have to select a "winning" parse
		and keep track of semantic values. Combining goto_target_node is the reason why
		semantic values go on edges, but this is only necessary for states reached by GOTO.
		Those reached by SHIFT might do as well to have a single semantic shared among
		all predecessors.
		
		Oh yes, one last thing: I seem to recall there's an ordering constraint.
		I think if you perform the shortest reductions first, the right things happen.
		Can this be encoded in the sequence of alternatives given?
		"""
		nonterminal_id, length, cid, view = table.get_rule(rule_id)
		for origin_node in node.all_paths(length):
			goto_state_id = table.get_goto(origin_node.state_id, nonterminal_id)
			if goto_state_id in top_of_stack:
				# At this point, if origin_node is already present, then local
				# ambiguity is indicated. For a recognizer, that is "packed" simply enough.
				top_of_stack[goto_state_id].edges.add(origin_node)
			else:
				goto_target_node = Node(goto_state_id, {origin_node})
				top_of_stack[goto_state_id] = goto_target_node
				queue.append(goto_target_node)
	
	initial_node = Node(table.get_initial(language), set())
	top_of_stack = {initial_node.state_id: initial_node}
	for symbol in sentence:
		shifts, queue = {}, list(top_of_stack.values())
		terminal_id = table.get_translation(symbol)
		for node in queue: act_on(node, table.get_action(node.state_id, terminal_id))
		if not shifts: raise interfaces.GeneralizedParseError("Parser died midway at something ungrammatical.")
		top_of_stack = shifts
	# Now deal with the end-of-input:
	shifts, queue = {}, list(top_of_stack.values())
	terminal_id = 0
	for node in queue: act_on(node, table.get_action(node.state_id, terminal_id))
	if not shifts: raise interfaces.GeneralizedParseError("Parser recognized a viable prefix, but not a complete sentence.")
	assert len(shifts) == 1
	return




"""
This module implements the macro-enabled enhanced-BNF context-free grammar semantics.
There is exactly zero concern for where the production rules come from.

Because the production rules support a no-kidding macro language, this module gets a
good bit of hair all by itself.

The MetaParse definition object works with structured rewriting-rule objects, so this
file begins with definitions of some semantic objects. Next come a grammar and scanner
(implemented via the "mini" infrastructure) which build those semantics up. Last is a
class definition for a grammar object which supplies the necessary bits to make
the extensions over BNF work properly.
"""

from ..support import failureprone, interfaces
from ..scanning import miniscan
from ..parsing import context_free, miniparse

NONDET = object()

class DefinitionError(Exception): pass

### Semantic categories for the breakdown of rewrite rules:
class Action:
	def __init__(self, name): self.name = name

class Element:
	"""
	Base class of RHS elements.
	What you do NOT see here is reflected in the constructor for class Rewrite:
		Specifically, any Element may have a 'capture' attribute added; this makes the value
		available to any action to its right.
		Please note: 'capturing' a rule-final action makes no sense and is presently ignored.
	"""
	def implement(self, ebnf:"EBNF_Definition", head:str, bindings:dict) -> str:
		""" Interpreting the MacroParse EBNF variant now seems to require a proper environment... """
		raise NotImplementedError(type(self))

class Symbol(Element):
	def __init__(self, name): self.name = name
	
	def implement(self, ebnf: "EBNF_Definition", head: str, bindings: dict) -> str:
		return head if self.name == '_' else bindings.get(self.name, self.name)


class InlineRenaming(Element):
	def __init__(self, alternatives):
		self.alternatives = tuple(alternatives)
		for a in self.alternatives: assert isinstance(a, Element) and not isinstance(a, Action)
	
	def implement(self, ebnf: "EBNF_Definition", head: str, bindings: dict) -> str:
		alts = [a.implement(ebnf, head, bindings) for a in self.alternatives]
		symbol = "[%s]"%("|".join(alts))
		if ebnf.implementing(symbol):
			for a in alts:
				ebnf.plain_cfg.rule(symbol, [a], None, None, 0, None)
		return symbol
		

class MacroCall(Element):
	def __init__(self, name, actual_parameters):
		self.name, self.actual_parameters = name, tuple(actual_parameters)
		assert isinstance(self.name, str)
		for a in self.actual_parameters: assert isinstance(a, Element), type(a)
	def canonical_symbol(self, bindings:dict) -> str: return
	
	def implement(self, ebnf: "EBNF_Definition", head: str, bindings: dict) -> str:
		args = [a.implement(ebnf, head, bindings) for a in self.actual_parameters]
		symbol = "%s(%s)"%(self.name, ",".join(args))
		if ebnf.implementing(symbol): ebnf.must_elaborate.append((symbol, self.name, args))
		return symbol


class Rewrite:
	def __init__(self, elements:list, precsym:str=None):
		self.elements = elements
		self.precsym = precsym
		self.line_nr = 0
		self.message:Action = self.elements.pop(-1) if isinstance(self.elements[-1], Action) else None
		self.__args = args = []
		self.size = len(self.elements)
		for i,elt in enumerate(self.elements):
			if hasattr(elt, 'capture'): args.append(i)
		if not args: args.extend(range(self.size)) # Pick up everything if nothing is special.
	
	def prefix_capture(self, size:int):
		return tuple(c - size for c in self.__args if c < size)
	
	def install(self, ebnf:"EBNF_Definition", head, bindings):
		ebnf.error_help.current_line_nr = self.line_nr # Because MACROS.
		rhs = [elt.implement(ebnf, head, bindings) for elt in self.elements]
		"""
		Install one rewrite rule:
		Does everything necessary to interpret extension forms down to plain BNF,
		and then enters that into self.plain_bnf as an option for `head`.
		"""
		raw_bnf = []
		for i, elt in enumerate(self.elements):
			if isinstance(elt, Action):
				placeholder = ':'+elt.name
				raw_bnf.append(placeholder)
				ebnf.internal_action(placeholder, self.prefix_capture(i))
			else:
				assert isinstance(elt, Element)
				raw_bnf.append(elt.implement(ebnf, head, bindings))
		
		if self.message is None:
			con = None
			plc = self.__args[0] if len(self.__args) == 1 else self.__args
		else:
			con, plc = self.message.name, self.__args
		ebnf.plain_cfg.rule(head, raw_bnf, self.precsym, con, plc, ebnf.error_help.current_line_nr)

"""
The MacroParse metagrammar contains various repetition constructs. The following functions
act as rudimentary macros to relieve some redundancy; they essentially mean something like:

	list_of(what, sep) -> what :first | .list_of(what, sep) sep .what :append
	one_or_more(what) -> what :first | .one_or_more(what) .what :append

Doing macros for real is a little more complicated and will be dealt with further down.
"""
def list_of(what, sep):
	head = 'list_of(%r,%r)'%(what,sep)
	METAGRAMMAR.rule(head, '.%s' % what)(FIRST)
	METAGRAMMAR.rule(head, '.%s %s .%s' % (head, sep, what))(APPEND)
	return head
def one_or_more(what):
	head = 'one_or_more(%r)' % (what)
	METAGRAMMAR.rule(head, '.%s' % what)(FIRST)
	METAGRAMMAR.rule(head, '.%s .%s' % (head, what))(APPEND)
	return head
def FIRST(element): return [element]
def APPEND(the_list, element):
	the_list.append(element)
	return the_list


### The MacroParse metagrammar for individual production lines is as follows.
METAGRAMMAR = miniparse.MiniParse('production', 'precedence', 'condition')

METAGRAMMAR.rule('production', '.head .' + list_of('rewrite', '|'))(None)

METAGRAMMAR.rule('head', '|')(None) # "use prior" is represented by a null "head".
METAGRAMMAR.rule('head', '.name arrow', )(None) # This will come across as a string.
METAGRAMMAR.rule('head', '.name ( .%s ) arrow' % list_of('name', ','))(None) # Represent macro heads with (name, args) tuples.

ELEMENTS = one_or_more('element')
METAGRAMMAR.rule('rewrite', ELEMENTS)(Rewrite)
METAGRAMMAR.rule('rewrite', '.' + ELEMENTS + ' pragma_prec .terminal')(Rewrite)

METAGRAMMAR.renaming('terminal', 'name', 'literal')

METAGRAMMAR.renaming('element', 'positional_element')
@METAGRAMMAR.rule('element', 'capture .positional_element')
def _capture_(positional_element:Element) -> Element:
	positional_element.capture = True
	return positional_element

METAGRAMMAR.renaming('positional_element', 'actual_parameter')
METAGRAMMAR.rule('positional_element', 'message')(Action)
METAGRAMMAR.rule('positional_element', 'err_token')(lambda x:Symbol(interfaces.ERROR_SYMBOL))

METAGRAMMAR.renaming('actual_parameter', 'symbol') # alternation brackets should not nest directly...
METAGRAMMAR.rule('actual_parameter', '[ .' + one_or_more('symbol') + ' ]')(InlineRenaming)

METAGRAMMAR.rule('symbol', 'name')(Symbol)    # Normal symbol
METAGRAMMAR.rule('symbol', 'literal')(Symbol) # Also a normal symbol with a funny name
METAGRAMMAR.rule('symbol', 'topic')(Symbol)   # Topic symbol; gets its name fixed later.
METAGRAMMAR.rule('symbol', '.name ( .' + list_of('actual_parameter', ',') + ' )')(MacroCall)

# Sub-language: For specifying precedence and associativity rules:
NAMES = one_or_more('name')
METAGRAMMAR.rule('precedence', '.associativity .'+one_or_more('terminal'))(None)
METAGRAMMAR.rule('precedence', 'pragma_nondeterministic .'+NAMES)(lambda x:(NONDET, x))

METAGRAMMAR.rule('associativity', 'pragma_left')(lambda x: context_free.LEFT)
METAGRAMMAR.rule('associativity', 'pragma_right')(lambda x: context_free.RIGHT)
METAGRAMMAR.rule('associativity', 'pragma_nonassoc')(lambda x: context_free.NONASSOC)
METAGRAMMAR.rule('associativity', 'pragma_bogus')(lambda x: context_free.BOGUS)

# Sub-language: For specifying the connections between scan conditions:
METAGRAMMAR.rule('condition', 'name')(lambda x:(x,[]))
METAGRAMMAR.rule('condition', '.name arrow .'+NAMES)(None)

### The lexeme definitions for production rule lines are as follows:
LEX = miniscan.Definition()
LEX.ignore(r'\s+') # Ignore whitespace
LEX.token('name', r'\l\w*') # Identifiers as token type "name".
LEX.token('err_token', r'\$error\$') # the error token is a bit special: it can't be an LHS or have precedence.
LEX.token('topic', r'_/\W') # Bare underline means "the current production rule head".
LEX.token_map('message', r':\l\w*', lambda tx:tx[1:]) # Strip out the colon for message names
@LEX.on(r'%\l+') # Build pragma token types from the text.
def pragma(yy): yy.token('pragma_'+yy.matched_text()[1:])
LEX.token('capture', r'[.]/\S') # a dot prefixes a captured element, so it needs to be followed by something.
@LEX.on(r'[][(),|]') # Punctuation is represented directly, with null semantic value.
def punctuate(yy): yy.token(yy.matched_text())
LEX.token_map('literal', r"'\S+'", lambda text:text[1:-1]) # Literals are allowed two ways, so you can
LEX.token_map('literal', r'"\S+"', lambda text:text[1:-1]) # easily contain whichever kind of quote.
LEX.token('arrow', r'[-=>:<]+') # Arrows in grammar definitions tend to look all different ways. This is flexible.

class ErrorHelper:
	"""
	Right now I'm not sure what else to call this. Naming things is hard, after all.
	The concept is to get the error reporting strategy defined once and accessible throughout.
	"""
	def __init__(self): self.current_line_nr = 0
	def gripe(self, message): raise DefinitionError('At line %d: %s.'%(self.current_line_nr, message))
	def gripe_about(self, line:str, column:int, message): self.gripe(message +"\n" + failureprone.illustration(line, column, prefix='\t'))
	def parse(self, line:str, line_nr:int, language:str):
		""" Factoring out the commonalities of half-decent error reporting... """
		assert isinstance(line_nr, int), type(line_nr)
		self.current_line_nr = line_nr
		metascan = LEX.scan(line)
		try: return METAGRAMMAR.parse(metascan, language=language)
		except interfaces.ScanError as e: self.gripe_about(line, e.args[0], "The MacroParse MetaScanner got confused by %r" % e.args[1])
		except interfaces.ParseError as e:
			self.gripe_about(
				line, metascan.current_position(),
				'The MacroParse MetaParser got confused about:\n\t'+e.condition()
			)


class MacroDefinition:
	def __init__(self, name:str, formals):
		self.name = name
		self.formals = tuple(formals)
		self.rewrites = []
		self.actually_used = False
	
	def elaborate(self, ebnf:"EBNF_Definition", head:str, args:list):
		if len(args) != len(self.formals):
			raise DefinitionError("Macro \"%s\" called with wrong number of arguments."%self.name)
		self.actually_used = True
		for rewrite in self.rewrites:
			rewrite.install(ebnf, head, dict(zip(self.formals, args)))

### Grammar object:
# Please note: The format of an action entry shall be the tuple <message_name, tuple-of-offsets, line_number>.
# These are a pass-through into the ContextFreeGrammar and eventually the parse tables themselves.

def illustrate_position(line:str, column:int): return line[:column]+'<<_here_>>'+line[column:]

class EBNF_Definition:
	def __init__(self, error_help:ErrorHelper):
		self.plain_cfg = context_free.ContextFreeGrammar()
		self.current_head = None # This bit of state facilitates the feature of beginning a line with an alternation symbol.
		self.inferential_start = None # Use this to infer a start symbol if necessary.
		self.macro_definitions = {} # name -> MacroDefinition
		self.implementations = {} # canonical symbol -> line number of first elaboration
		self.must_elaborate = []
		self.error_help = error_help
		self.nondeterministic_symbols = set()
	
	def read_precedence_line(self, line:str, line_nr:int):
		direction, symbols = self.error_help.parse(line, line_nr, 'precedence')
		if direction is NONDET: self.nondeterministic_symbols.update(symbols)
		else: self.plain_cfg.assoc(direction, symbols)

	def read_production_line(self, line:str, line_nr:int):
		""" This is the main interface to defining grammars. Call this repeatedly for each line of the grammar. """
		head, rewrites = self.error_help.parse(line, line_nr, 'production')
		for R in rewrites: R.line_nr = line_nr
		# Set the current head field, or use it unchanged if not specified on this line:
		if head is None:
			if self.current_head is None:
				self.error_help.gripe('Confused about what the current head nonterminal is')
			head = self.current_head
		else:
			if isinstance(head, tuple): # Do we need to enter a macro declaration?
				name, formals = head
				if name in self.macro_definitions: # Prevent re-declarations.
					self.error_help.gripe("Cannot re-declare macro %r, which was orginally declared on line %d."%(name, self.macro_definitions[name].line_nr))
				elif len(set(formals)|{name}) <= len(formals):
					self.error_help.gripe("All the names used in a macro head declaration must be distinct.")
				else:
					head = self.macro_definitions[name] = MacroDefinition(name, formals)
			self.current_head = head
		# Proceed to do the right thing with supplied rewrite rules on this line:
		if isinstance(self.current_head, MacroDefinition):
			self.current_head.rewrites.extend(rewrites)
		else:
			assert isinstance(head, str)
			if self.inferential_start is None: self.inferential_start = head
			for R in rewrites:
				R.install(self, head, {})
		pass
	
	def validate(self):
		unused_macros = sorted(name+' at line '+str(definition.rewrites[0].line_nr) for name, definition in self.macro_definitions.items() if not definition.actually_used)
		if unused_macros: raise DefinitionError('The following macro(s) were defined but never used:\n\t'+'\n\t'.join(unused_macros))
		if not self.inferential_start: raise DefinitionError("No production rules have been given, so how am I to compile a grammar? (You could give a trivial one...)")
		if not self.plain_cfg.start:
			print('Inferring CFG start symbol %r from earliest production because none was given explicitly.'%self.inferential_start)
			self.plain_cfg.start.append(self.inferential_start)
		self.plain_cfg.validate()
	
	def sugarless_form(self) -> context_free.ContextFreeGrammar:
		if self.inferential_start: # In other words, if any rules were ever given...
			while self.must_elaborate:
				(symbol, name, args) = self.must_elaborate.pop()
				try: definition = self.macro_definitions[name]
				except KeyError: raise DefinitionError("At line %d macro \"%s\" is called but never defined."%(self.implementations[symbol], name))
				else: definition.elaborate(self, symbol, args)
			self.validate()
			return self.plain_cfg
	
	def internal_action(self, placeholder:str, capture:tuple):
		""" These are implemented internally much like an epsilon rule, but able to capture certain left-context. """
		if self.implementing(placeholder):
			self.plain_cfg.rule(placeholder, [], None, placeholder, capture, self.error_help.current_line_nr)
		else:
			self.error_help.gripe(
				'Internal action %s was first elaborated on line %d; reuse is not (presently) supported.' %
				(placeholder, self.implementations[placeholder])
			)

	def implementing(self, symbol) -> bool:
		""" Tells if the current instantiation of a macro-rule is the first; notes the line number where it happened. """
		if symbol in self.implementations:
			return False
		else:
			self.implementations[symbol] = self.error_help.current_line_nr
			return True



""" Small is beautiful. These algorithms need no introduction. """

from collections.abc import Sequence, Mapping, Iterable
import operator

def allocate(a_list:list, item):
	""" Too frequent an idiom not to abbreviate. """
	idx = len(a_list)
	a_list.append(item)
	return idx

def transitive_closure(roots, successors) -> set:
	""" Transitive closure is a simple unordered graph exploration. """
	black = set()
	grey = set(roots)
	while grey:
		k = grey.pop()
		if k not in black:
			black.add(k)
			them = successors(k)
			if them is not None: grey.update(them)
	return black

class BreadthFirstTraversal:
	""" This object also accumulates and exposes data about the traversal path. """
	def __init__(self):
		self.current, self.traversal, self.catalog, self.earliest_predecessor, self.breadcrumbs = None, [], {}, [], []
	def execute(self, visit):
		""" visit(key, lookup) should call self.lookup(successor_key, breadcrumb), which returns an integer. """
		for self.current, key in enumerate(self.traversal): visit(key)
		self.current = None
	def lookup(self, key, *, breadcrumb=None) -> int:
		if key not in self.catalog:
			self.catalog[key] = allocate(self.traversal, key)
			self.earliest_predecessor.append(self.current)
			self.breadcrumbs.append(breadcrumb)
		return self.catalog[key]
	def depth_list(self) -> list:
		result = []
		for p in self.earliest_predecessor: result.append(0 if p is None else 1+result[p])
		return result
	def shortest_path_to(self, index:int) -> list:
		""" Return a minimal list of states traversed, from a root to the given node index, in normal order. """
		path = []
		while index is not None:
			path.append(index)
			index = self.earliest_predecessor[index]
		path.reverse()
		return path

class EquivalenceClassifier:
	def __init__(self):
		self.catalog = {}
		self.exemplars = []
	def classify(self, key):
		if key not in self.catalog:
			self.catalog[key] = allocate(self.exemplars, key)
		return self.catalog[key]

def hamming_distance(a, b):
	""" Compute the number of places in two (equal-length) sequences with unequal corresponding elements. """
	assert len(a) == len(b)
	return sum(map(operator.ne, a, b))

def grade(seq:Sequence, *, descending=False) -> list:
	""" returns result such that [seq[i] for i in grade(seq)] == sorted(seq). Cribbed from APL's grade_up/down. """
	return sorted(range(len(seq)), key=seq.__getitem__, reverse=descending)

def everted(permutation:Sequence) -> list:
	""" Creates and returns an inverse permutation: like it's been turned inside-out """
	result = [None] * len(permutation)
	for i, x in enumerate(permutation): result[x] = i
	return result

def collation(seq:Sequence, *, descending=False):
	""" The inverse of `grade`: Tells you how a sorted sequence was permuted to get here. """
	return everted(grade(seq, descending=descending))

def strongly_connected_components_by_tarjan(graph):
	"""
	See https://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
	Returns a list of strongly-connected components in reverse topological order.
	Each component is a list of member node numbers. Deviating slightly from the wikipedia
	presentation, the low-link is local to the recursive call, eliminating one confusion.
	
	It's expected that graph[q] is the list of arcs from (or perhaps to) node q.
	The linear-time bound assumes all nodes are numbered from 0..last. An isomorphism
	for hashable node keys is provided below.
	"""
	def unvisited(q): return index[q] is None
	def connect(q) -> int:
		low_link = index[q] = allocate(stack, q)
		on_stack[q] = True
		for r in graph[q]:
			if unvisited(r): low_link = min(low_link, connect(r))
			elif on_stack[r]: low_link = min(low_link, index[r])
		if low_link == index[q]: # i.e. if node q is the root of an SCC:
			component = stack[low_link:]
			del stack[low_link:]
			for r in component: on_stack[r] = False
			output.append(component)
		return low_link
	def main():
		for q in range(size):
			if unvisited(q): connect(q)
	size = len(graph)
	index = [None] * size
	on_stack = [False] * size
	stack = []
	output = []
	main()
	return output

def strongly_connected_components_hashable(graph:dict):
	""" This adapts Tarjan's SCC algorithm for more kinds of graph node labels than strictly integers. """
	table = list(graph.keys())
	index = {key:i for i,key in enumerate(table)}
	prime = [
		[index[arc] for arc in node if arc in index]
		for node in graph.values()
	]
	return [[table[q] for q in component] for component in strongly_connected_components_by_tarjan(prime)]



""" Mechanisms for working with regular languages: finite automata and an AST class-hierarchy for regular expressions. """

import typing, operator

from ..support import foundation, pretty, interfaces
from ..scanning import charclass, charset


class DFA(interfaces.FiniteAutomaton):
	"""  """
	def __init__(self, *, alphabet: interfaces.Classifier, initial:dict, final:dict, states:list):
		self.alphabet = alphabet
		self.width = self.alphabet.cardinality()
		self.initial = initial
		self.final = final
		metric = [len(row) for row in states]
		assert all(m==self.width for m in metric), (self.width, metric)
		self.states = states
	def jam_state(self): return -1
	def append_state(self, row) -> int:
		assert len(row) == self.width, [len(row), self.width]
		return foundation.allocate(self.states, row)
	def get_condition(self, condition_name) -> tuple: return self.initial[condition_name]
	def get_next_state(self, current_state: int, codepoint: int) -> int: return self.states[current_state][self.alphabet.classify(codepoint)]
	def get_state_rule_id(self, state_id: int): return self.final.get(state_id)
	
	def display(self):
		print('Finite Automaton:')
		self.alphabet.display()
		print('Initial:', self.initial)
		head = ['*', '', ]+list(range(self.alphabet.cardinality()))
		body = [[self.final.get(i,''), i, *[s if s >= 0 else '' for s in row]] for i, row in enumerate(self.states)]
		pretty.print_grid([head] + body)
	
	def minimize_states(self) -> "DFA":
		""" Let's try Moore's Algorithm: It's easier to get right. """
		buckets = []
		partition_id_of_state = []
		def establish_initial_partitioning():
			""" Every state goes into a partition numbered according to its finality: """
			finality = {}
			for n in range(len(self.states)):
				rule_id = self.final.get(n)
				if rule_id not in finality: finality[rule_id] = foundation.allocate(buckets, [])
				b = finality[rule_id]
				partition_id_of_state.append(b)
				buckets[b].append(n)
			pass
		def translate(q) -> tuple: return tuple(-1 if x < 0 else partition_id_of_state[x] for x in self.states[q])
		def refine_partitions() -> bool:
			splitting_happened = False
			for b, bucket in enumerate(buckets):
				if len(bucket) < 2: continue
				exemplar = translate(bucket[0])
				same, different = [], []
				for q in bucket: (same if all(map(operator.eq, exemplar, translate(q))) else different).append(q)
				if different:
					buckets[b] = same
					n = foundation.allocate(buckets, different)
					for q in different: partition_id_of_state[q] = n
					splitting_happened = True
			return splitting_happened
		def interpretation() -> DFA:
			return DFA(
				alphabet=self.alphabet,
				initial={condition: (partition_id_of_state[q0], partition_id_of_state[q1]) for condition, (q0, q1) in self.initial.items()},
				final = {partition_id_of_state[q]:rule_id for q,rule_id in self.final.items()},
				states = [translate(bucket[0]) for bucket in buckets],
			)
		establish_initial_partitioning()
		while refine_partitions(): pass
		return interpretation()
		
	def minimize_alphabet(self) -> "DFA":
		assert isinstance(self.alphabet, charclass.SimpleClassifier)
		ec = foundation.EquivalenceClassifier()
		classes = [ec.classify(column) for column in zip(*self.states)]
		return DFA(
			alphabet=charclass.MetaClassifier(self.alphabet.bounds, classes),
			initial=self.initial,
			final=self.final,
			states=list(zip(*ec.exemplars)),
		)
	
	def stats(self):
		Q = len(self.states)
		W = self.width
		X = sum(sum(x != -1 for x in state) for state in self.states)
		print('DFA has %d states and %d character classes, using %d cells.'%(Q, W, Q*W))
		print('%d non-error cells, or %0.2f%%'%(X, 100*X/(Q*W)))
	
	def make_csv(self, pathstem):
		pretty.write_csv_grid(pathstem + '.dfa.csv', [
			[q, self.final.get(q), '', *row]
			for q, row in enumerate(self.states)
		])
		pass

class NFA:
	def __init__(self):
		self.states, self.initial, self.final = [], {}, {}
		self.all_bounds = {0} # There is at least an "end-of-file" marker, which is distinct from all else.

	class Edge(typing.NamedTuple):
		label: list
		target: int

	class Node(typing.NamedTuple):
		edges: list
		epsilons: set
		rank: int
	
	def new_node(self, rank) -> int: return foundation.allocate(self.states, NFA.Node(edges=[], epsilons=set(), rank=rank))
	def condition(self, name):
		if name not in self.initial: self.initial[name] = (self.new_node(0), self.new_node(0))
		return self.initial[name]
	def link(self, src :int, dst :int, label :list):
		self.states[src].edges.append(NFA.Edge(label, dst))
		self.all_bounds.update(label)
	def link_epsilon(self, src :int, dst :int): self.states[src].epsilons.add(dst)
	def link_condition(self, main_condition, included_condition):
		"""
		Precondition: Both conditions exist, and are known to the NFA `self`.
		Postcondition: It is as if all the rules in `included_condition` are also defined in `main_condition`,
		and in the same relative order.
		"""
		for src, dst in zip(self.initial[main_condition], self.initial[included_condition]):
			self.link_epsilon(src, dst)
	def subset_construction(self) -> DFA:
		"""
		The standard plan to convert an NFA to a DFA, in plain English, is to consider
		that a deterministic state represents a particular and distinct subset of NFA states.
		That general plan is embodied by the interplay between the generic BreadthFirstTraversal(...)
		and `visit(key)`: the "key" is composed principally of a `frozenset` of NFA state ID numbers.
		
		This module also supports "rule ranks" -- about which more is written elsewhere. To implement
		the concept, an extra bit of data rides along with the subset: specifically, a rank number.
		"""
		def close(ns, min_rank:int):
			assert all(isinstance(n, int) for n in ns)
			tc = foundation.transitive_closure(ns, lambda n :self.states[n].epsilons)
			ns = frozenset(n for n in tc if self.states[n].rank >= min_rank)
			return ns, min((self.states[n].rank for n in ns), default=0)
		def visit(key):
			ns, min_rank = key
			active, targets = [], []
			final = ns & self.final.keys()
			if final:
				min_rank = max(self.states[n].rank for n in final)
				dfa.final[len(dfa.states)] = min(self.final[n] for n in final if self.states[n].rank == min_rank)
			for n in ns:
				node = self.states[n]
				if node.rank >= min_rank:
					for e in node.edges:
						active.append(list(charset.expand(e.label, all_bounds)))
						targets.append(e.target)
			# Now, if active[i][j], then targets[i] participates in class[j].
			delta, prior = [], None
			for column in zip(*active):
				register = [t for r,t in zip(column, targets) if r]
				if register != prior: # Save a lot of transitive closure operations...
					prior = register
					subset, subrank = close(register, min_rank)
					successor = bft.lookup((subset, subrank)) if subset else -1
				delta.append(successor)
			dfa.append_state(delta or [-1]*dfa.width)
		# Main routine:
		all_bounds = sorted(self.all_bounds|{-1})
		dfa = DFA(alphabet=charclass.SimpleClassifier(all_bounds[1:]), initial={}, final={}, states=[])
		bft = foundation.BreadthFirstTraversal()
		def initial(q:int): return close([q], 0)
		dfa.initial = {k :(bft.lookup(initial(a)), bft.lookup(initial(b))) for k ,(a ,b) in self.initial.items()}
		bft.execute(visit)
		return dfa

class Regular:
	def encode(self, src:int, dst:int, nfa:NFA, rank): raise NotImplementedError(type(self))
	def length(self): raise NotImplementedError(type(self)) # Can't use __len__ because Python runtime demands an integer.

class CharClass(Regular):
	def __init__(self, cls:list): self.cls = cls
	def encode(self, src:int, dst:int, nfa:NFA, rank): nfa.link(src, dst, self.cls)
	def length(self): return 1

class Alternation(Regular):
	def __init__(self, a:Regular, b:Regular): self.a, self.b = a,b
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		self.a.encode(src, dst, nfa, rank)
		self.b.encode(src, dst, nfa, rank)
	def length(self):
		a, b = self.a.length(), self.b.length()
		if a==b: return a

class Sequence(Regular):
	def __init__(self, a:Regular, b:Regular): self.a, self.b = a,b
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		midpoint = nfa.new_node(rank)
		self.a.encode(src, midpoint, nfa, rank)
		self.b.encode(midpoint, dst, nfa, rank)
	def length(self):
		a,b = self.a.length(), self.b.length()
		if None not in (a,b): return a+b

class Inflection(Regular):
	def __init__(self, sub:Regular): self.sub = sub
	def length(self): return None

class Star(Inflection):
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		loop = nfa.new_node(rank)
		self.sub.encode(loop, loop, nfa, rank)
		nfa.link_epsilon(src, loop)
		nfa.link_epsilon(loop, dst)

class Hook(Inflection):
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		self.sub.encode(src, dst, nfa, rank)
		nfa.link_epsilon(src, dst)

class Plus(Inflection):
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		before, after = nfa.new_node(rank), nfa.new_node(rank)
		self.sub.encode(before, after, nfa, rank)
		nfa.link_epsilon(src, before)
		nfa.link_epsilon(after, before)
		nfa.link_epsilon(after, dst)

class Counted(Regular):
	""" Just too handy not to provide. """
	def __init__(self, sub: Regular, m:int, n):
		assert isinstance(sub, Regular)
		assert isinstance(m, int)
		assert isinstance(n, int) or n is None
		self.sub, self.m, self.n = sub, m, n
	def encode(self, src:int, dst:int, nfa:NFA, rank):
		p1 = nfa.new_node(rank)
		nfa.link_epsilon(src, p1)
		for i in range(self.m):
			p2 = nfa.new_node(rank)
			self.sub.encode(p1, p2, nfa, rank)
			p1 = p2
		nfa.link_epsilon(p1, dst)
		if self.n is None:
			self.sub.encode(p1, p1, nfa,rank)
		else:
			for i in range(self.n - self.m):
				p2 = nfa.new_node(rank)
				self.sub.encode(p1, p2, nfa, rank)
				nfa.link_epsilon(p2, dst)
				p1 = p2
	def length(self):
		if self.m == self.n:
			x = self.sub.length()
			if x is not None: return x * self.m

		



"""
This
"""

import bisect
from ..support import pretty, interfaces


class SimpleClassifier(interfaces.Classifier):
	def __init__(self, bounds):
		self.bounds = tuple(bounds)
	def cardinality(self): return 1+len(self.bounds)
	def classify(self, codepoint:int): return bisect.bisect_right(self.bounds, codepoint)
	def display(self): pretty.print_grid([['-', *self.bounds]])

class MetaClassifier(interfaces.Classifier):
	def __init__(self, bounds, classes):
		self.bounds = tuple(bounds)
		self.classes = tuple(classes)
		assert len(self.bounds) + 1 == len(self.classes)
		assert set(self.classes) == set(range(self.cardinality()))
	def cardinality(self): return max(self.classes)+1
	def classify(self, codepoint): return self.classes[bisect.bisect_right(self.bounds, codepoint)]
	def display(self):
		bound_repr = ["'"+chr(b) if 32 < b < 127 else b for b in self.bounds]
		pretty.print_grid([('-', *bound_repr), self.classes, ])



"""
Discarded ideas go here: they may be educational but they are not the current direction of the project.

This module is quite deliberately not maintained.

Then again, project direction has changed before.
"""
from .support import foundation
from .support.compaction import most_common, is_homogeneous

def multi_append(table:dict, row:dict):
	"""
	Utility function for collections of parallel arrays; these often have better
	storage and encoding characteristics than arrays of records do.
	:param table: A dictionary in which several keys point to lists.
	:param row: A dictionary; the values get appended to the corresponding lists in `table`.
	:return: Nothing.
	"""
	for key, value in row.items(): table[key].append(value)


def modified_aho_corasick_encoding(*, initial: dict, matrix: list, final: dict, jam: int) -> dict:
	"""
	Alfred V. Aho and Margaret J. Corasick discovered a truly wonderful algorithm for a particular class
	of string search problem in which there are both many needles and lots of haystack data. It works by
	the construction of a TRIE for identifying needles, together with a failure pointer for what state to
	enter if the present haystack character is not among the outbound edges from the current state. The
	structure so constructed is queried at most twice per character of haystack data on average, because
	the failure pointers always point to a node at least one edge less deep into the TRIE.

	That structure provides the inspiration for a fast, compact encoding of an arbitrary DFA. The key
	idea is to find, for each state, the least-cost "default transition", defined according to how much
	storage gets used under the rules for that case. We begin with the most common transition within
	the row, but then scan over prior states looking for a good template: a "shallower" state with
	many transitions in common. It is then sufficient to store the identity of that default pointer
	and a sparse list of exceptions. If the sparse list results in expansion, a dense row may be stored
	instead. If the exceptions are all identical, they may be stored as a single integer rather than a list.

	Upon further reflection: sometime states at the same distance-from-root have the most in common.
	A probe takes constant amortized time as long as the fail-over path length is restricted to a constant
	multiple of the node's depth. That constant probably need not be more than one for excellent results.
	"""
	# Renumber the states according to a breadth-first topology, and also determine the resulting
	# depth boundaries. (A topological index would suffice, but this has other nice properties.)
	assert jam == -1
	bft = foundation.BreadthFirstTraversal()
	states = []
	
	def renumber(src): states.append([jam if dst == jam else bft.lookup(dst) for dst in matrix[src]])
	
	initial = {condition: (bft.lookup(q0), bft.lookup(q1)) for condition, (q0, q1) in initial.items()}
	bft.execute(renumber)
	depth = bft.depth_list()
	
	# Next, we need to construct the compressed structure. For simplicity, this will be three arrays
	# named for their function and indexed by state number.
	delta = {'default': [], 'index': [], 'data': [], }
	result = {'delta': delta, 'initial': initial, 'final': [bft.lookup(q) for q in final.keys()],
	          'rule': list(final.values())}
	failover_path_length = []
	metric = 0
	for i, row in enumerate(states):
		# Find the shortest encoding, possibly by reference to earlier states:
		default = most_common(row)
		index = [k for k, x in enumerate(row) if x != default]
		data = [row[k] for k in index]
		cost = len(index) + (1 if is_homogeneous(data) else len(data))
		for j in range(i):
			if failover_path_length[j] <= depth[i]:
				try_index = [k for k, x in enumerate(states[j]) if x != row[k]]
				try_data = [row[k] for k in try_index]
				try_cost = len(try_index) + (1 if is_homogeneous(try_data) else len(try_data))
				if try_cost < cost: default, index, data, cost = -2 - j, try_index, try_data, try_cost
		# Append the chosen encoding into the structure:
		if cost < len(row):  # If the compression actually saves space on this row:
			if is_homogeneous(data): data = data[0] if data else default
			multi_append(delta, {'default': default, 'index': index, 'data': data})
			failover_path_length.append(0 if default > -2 else 1 + failover_path_length[-2 - default])
		else:  # Otherwise, a dense storage format is indicated by eliding the list of indices.
			multi_append(delta, {'default': jam, 'index': None, 'data': row})
			cost = len(row)
			failover_path_length.append(0)
		metric += 1 + cost
	# At this point, the DFA is represented in about as terse a format as makes sense.
	raw_size = len(matrix) * len(matrix[0])
	print('Matrix compressed into %d cells (%0.2f%% of %d * %d = %d) with at most %d probes per character scanned.' % (
	metric, 100 * metric / raw_size, len(matrix), len(matrix[0]), raw_size, 1 + max(failover_path_length)))
	return result

def aho_corasick_style_scanner_delta_function(*, index, data, default) -> callable:
	""" This decodes the above compaction by following `default` pointers as needed until an answer is found. """
	probe = sparse_table_function(index=index, data=data)
	def fn(state_id:int, symbol_id:int) -> int:
		if state_id<0: return state_id
		q = probe(state_id, symbol_id)
		if q is None:
			d = default[state_id]
			return d if d > -2 else fn(-2-d, symbol_id)
		else: return q
	return fn

def old_parser_action_function(*, index, data, default) -> callable:
	""" This part adds the "default reductions" layer atop a naive representation of a sparse action table. """
	probe = sparse_table_function(index=index, data=data)
	def fn(state_id:int, symbol_id:int) -> int:
		q = probe(state_id, symbol_id)
		return default[state_id] if q is None else q
	return fn

def old_parser_goto_function(*, state_class, class_list ) -> callable:
	def probe(state_id:int, nonterminal_id:int):
		cls = state_class[state_id]
		return 0-cls if cls < 0 else class_list[cls][nonterminal_id]
	return probe


def sparse_table_function(*, index, data) -> callable:
	"""
	The very simplest Python-ish "sparse matrix", and plenty fast on modern hardware, for the
	size of tables this module will probably ever see, is an ordinary Python dictionary from
	<row,column> tuples to significant table entries. There are better ways if you program
	closer to the bare metal, but this serves the purpose.

	This routine unpacks "compressed-sparse-row"-style data into an equivalent Python dictionary,
	then returns a means to query said dictionary according to the expected 2-dimensional interface.
	"""
	hashmap = {}
	for row_id, (Cs, Ds) in enumerate(zip(index, data)):
		if isinstance(Ds, int):  # All non-blank cells this row have the same value:
			for column_id in Cs: hashmap[row_id, column_id] = Ds
		else:
			for column_id, d in zip(Cs, Ds) if Cs else enumerate(Ds):
				hashmap[row_id, column_id] = d
	return lambda R, C: hashmap.get((R, C))


def compress_action_table(matrix: list) -> dict:
	"""
	Produce a compact representation of the "ACTION" table for a typical shift-reduce parser.
	:param matrix: list-of-lists of parse actions: positive numbers are shifts; negative are reductions, zero is error.
	:return: a compact structure.

	For each state in the ACTION table, the reduction with the largest lookahead set becomes a
	default reduction, so that only exceptions need be stored. This can delay the detection of
	a syntax error until several reductions have taken place, but it does not change the set
	of strings accepted as part of the language. Also, unless measures are taken to track the
	deepest state before the last round of reductions, it can throw away information useful to
	compute a set of valid expected next tokens in an error condition.

	Two subtleties affect the above strategy:
	1. Non-associativity declarations can result in	cells which MUST error despite being in
	   states which otherwise may have a default reduction. These can be expressly included in
	   the list of exceptions to the default.
	2. Error productions result in states that shift on the ERROR token. These states should
	   not get default-reductions -- or if they do, then the above-styled special measures
	   MUST be taken to ensure that the most suitable error production is finally used.

	This parsing library doesn't yet support error productions, so I'm only going to worry
	about the first case for now. Error productions might be a nice enhancement.

	The encoding of actions here is the same as is used in automata.DragonBookTable.
	"""
	
	def find_default_reduction(row: list) -> int:
		rs = [x for x in row if x < 0]
		return most_common(rs) if rs else 0
	
	raw_size = len(matrix) * len(matrix[0])
	print('Action table begins with %d states and %d columns (%d cells).' % (len(matrix), len(matrix[0]), raw_size))
	metric = 0
	delta = {'default': [], 'index': [], 'data': []}
	for q, row in enumerate(matrix):
		reduction = find_default_reduction(row)
		becomes_default = (reduction, 0)
		idx = [i for i, x in enumerate(row) if x not in becomes_default and (q, x) not in essential_errors]
		if len(idx) * 2 < len(row):
			multi_append(delta, {'default': reduction, 'index': idx, 'data': [row[k] for k in idx]})
			metric += 3 + 2 * len(idx)
		else:
			multi_append(delta, {'default': 0, 'index': None, 'data': row})
			metric += 2 + len(row)
	print("Compact form takes %d cells (%0.2f%%)." % (metric, 100 * metric / raw_size))
	return delta




"""
G-for-Generalized: GLR is basically the non-deterministic version of LR parsing.

As a parsing strategy, GLR means non-determinism is embraced and resolved as necessary at
in the context of actual parse runs. You can do GLR with any viable LR-style HFA, though
better tables naturally result in fewer blind alleys. In principle, the "trivial" table
is just the set of all grammar rules at every point: this yields classical Earley parsing.

Within this module, various handle-finding automaton (HFA) construction functions return a
(mostly) standardized structure which properly represents any remaining non-determinism.
Transmuting these to fully-deterministic structures is the LR module's job.

The HFA class itself contains a trial-parse routine which exercises these constructions
by determining whether they recognize a string as "in the language". It's great for testing
and illustrates one method to perform a parallel-parse, but it does not bother to recover a
parse tree or semantic value. Such things could be added, but preferably atop good means
for persisting ambiguous parse tables.
"""
import collections, sys
from typing import NamedTuple, Iterable, TypeVar, Generic, List, Dict, Set, Tuple
from ..support import foundation, pretty, interfaces
from . import context_free

T = TypeVar('T')

class PurityError(ValueError):
	""" Raised if a grammar has the wrong/undeclared conflicts. """

class HFA(Generic[T]):
	"""
	HFA is short for "Handle-Finding Automaton", which is clunky to repeat all over the place.
	
	What is a handle?
	
	A handle is that point in the text where the right end of a rule has been matched and
	recognition of the corresponding non-terminal symbol could be performed.
	
	There is a common base structure for these, and certain operations are largely similar
	regardless of which sort of table-construction algorithm is underway.
	
	The main difference between HFAs is what sort of object they use for states.
	To participate in the generic operations, states must have a "shift" dictionary mapping
	symbols to successor state numbers (not objects) and support state.reductions_before(lexeme)
	which must return an Iterable of zero or more rule-IDs.
	
	Fields are:
	graph: a list of state objects; their index is implicitly their node ID.
	initial: a list of initial-state ID numbers (graph node indices) corresponding
		to the start-symbols of the CFG.
	accept: a list of final/accepting-state ID numbers (graph node indices) also
		corresponding to the start-symbols of the CFG.
	grammar: The ContextFreeGrammar that started it all.
	bft: The BreadthFirstTraversal object which was used for the construction.
		This happens to be greatly useful in various diagnostic and other capacities.
	"""
	graph: List[T]
	initial: List[int]
	accept: List[int]
	grammar: context_free.ContextFreeGrammar
	bft: foundation.BreadthFirstTraversal
	
	def __init__(self, *, graph, initial, accept, grammar, bft):
		""" I don't want exactly named-tuple semantics. """
		self.graph, self.initial,self.accept, self.grammar, self.bft = graph, initial, accept, grammar, bft
	
	def display_situation(self, q: int, lookahead: str):
		"""
		Used for diagnostic displays:
			How might I get to state #q, in symbols?
			What would that parser situation look like?
		"""
		head, *tail = self.bft.shortest_path_to(q)
		print('==============\nIn language %r, consider:' % self.grammar.start[self.initial.index(head)])
		print('\t' + ' '.join(map(self.bft.breadcrumbs.__getitem__, tail)), pretty.DOT, lookahead)
	
	def earley_core(self, q:int):
		""" Maybe we need to know the core-items that a state began from, stripped of look-ahead powers. """
		return sorted(set((r,p) for r, p, *_ in self.bft.traversal[q]))
	
	def traverse(self, q: int, symbols:Iterable) -> int:
		""" Starting in state q, follow the shifts for symbols, and return the resulting state ID. """
		for s in symbols: q = self.graph[q].shift[s]
		return q
	
	def make_dot_file(self, path):
		""" Make a file suitable for the "dot" application from the Graphviz package. """
		with open(path, 'w') as fh:
			fh.write("digraph {\n")
			for q, state in enumerate(self.graph):
				sym = self.bft.breadcrumbs[q] or ''
				sym = sym.replace('"', r'\"')
				if sym.endswith('\\'): sym = sym + ' '
				fh.write("%d [label=\"%d: %s\"]\n"%(q, q, sym))
				for i in state.shift.values():
					fh.write("\t%d -> %d\n"%(q,i))
			fh.write('}\n')
		pass
	
	def trial_parse(self, sentence: Iterable):
		"""
		This is intended to be a super-simplistic non-deterministic recognizer: It exists only for
		unit-testing the GLR table-construction algorithms, and therefore doesn't try to build a
		semantic value or worry about pathological grammars.
		
		The approach taken is a lock-step parallel simulation with a cactus-stack of viable states:
		each entry is a cons cell consisting of a state id and prior stack. Because it explores
		every possible parse, it will diverge if faced with an infinitely-ambiguous situation.
		There are ways to cope with such cases, but in practice they are normally the result of
		mistakes, so the more useful response is to reject infinitely-ambiguous grammars.

		To play along, HFA states must support the .reductions_before(lexeme) method.
		"""
		
		language_index = 0 # Or perhaps: self.grammar.start.index[language_symbol]
		initial, accept = self.initial[language_index], self.accept[language_index]
		
		def reduce(stack, rule_id):
			""" To perform a reduction, roll the stack to before the RHS and then shift the LHS. """
			rule = self.grammar.rules[rule_id]
			for i in range(len(rule.rhs)): stack = stack[1]
			return self.graph[stack[0]].shift[rule.lhs], stack
		
		root = (initial, None)
		alive = [root]
		for lexeme in sentence:
			next = []
			for stack in alive:
				state = self.graph[stack[0]]
				if lexeme in state.shift: next.append((state.shift[lexeme], stack))
				for rule_id in state.reductions_before(lexeme): alive.append(reduce(stack, rule_id))
			alive = next
			if not alive: raise interfaces.GeneralizedParseError("Parser died midway at something ungrammatical.")
		for stack in alive:
			q = stack[0]
			if q == accept: return True
			for rule_id in self.graph[q].reductions_before(interfaces.END_OF_TOKENS):
				alive.append(reduce(stack, rule_id))
		raise interfaces.GeneralizedParseError("Parser recognized a viable prefix, but not a complete sentence.")

class LR0_State(NamedTuple):
	"""
	The LR(0) construction completely ignores look-ahead for reduce-rules, so for a
	non-deterministic parse table, it's enough to track a set of possible rules.
	"""
	shift: Dict[str, int]  # symbol => state-id
	reduce: Set[int]  # rule-id
	def reductions_before(self, lexeme): return self.reduce


def lr0_construction(grammar: context_free.ContextFreeGrammar) -> HFA[LR0_State]:
	"""
	In broad strokes, this is a subset-construction with a sophisticated means
	to identify successor-states. The keys (by which nodes are identified) are
	core-sets of LR0 parse items, themselves pairs of (rule-id, position).
	
	Additionally, during the full-elaboration step in visiting a core, completed
	parse-items correspond to a state's `reduce` entries. We don't worry about
	look-ahead in this construction: hence the '0' in LR(0). The net result is
	a compact table generated very quickly, but with somewhat limited power.
	In practical systems, LR(0) is normally just a first step, but some few
	grammars are deterministic in LR(0).
	"""
	
	def build_state(core: frozenset):
		step, reduce = collections.defaultdict(set), set()
		def visit(item):
			rule_id, position = item
			if position < len(RHS[rule_id]):
				next_symbol = RHS[rule_id][position]
				step[next_symbol].add((rule_id,position+1))
				return symbol_front.get(next_symbol)
			elif rule_id<len(grammar.rules): reduce.add(rule_id)
		foundation.transitive_closure(core, visit)
		graph.append(LR0_State(shift=ure.find_shifts(step), reduce=reduce, ))
	
	assert grammar.start
	##### Arranging certain bits of the grammar data in a convenient form:
	RHS = grammar.augmented_rules()
	def front(rule_ids): return frozenset([(r,0) for r in rule_ids])
	symbol_front = {symbol: front(rule_ids) for symbol, rule_ids in grammar.symbol_rule_ids.items()}
	bft = foundation.BreadthFirstTraversal()
	ure = UnitReductionEliminator(grammar, bft)
	graph = []
	initial = [bft.lookup(front([rule_id])) for rule_id in grammar.initial()]
	bft.execute(build_state)
	accept = [graph[qi].shift[language] for qi, language in zip(initial, grammar.start)]
	return HFA(graph=graph, initial=initial, accept=accept, grammar=grammar, bft=bft)


class UnitReductionEliminator:
	def __init__(self, grammar: context_free.ContextFreeGrammar, bft: foundation.BreadthFirstTraversal):
		self.bft = bft
		self.unit_rules = {}
		self.eligible_rhs = set()
		for rule_id, rule in enumerate(grammar.rules):
			if rule.is_rename():
				self.unit_rules[rule_id] = rule.lhs
				self.eligible_rhs.add(rule.rhs[0])

	def find_shifts(self, step: dict) -> dict:
		"""
		A unit-rule is eligible to be elided (optimized-to-nothing) exactly when
		the state reached by shifting the RHS would have no parse items associated with
		any other rule. Naively, we'd shift the RHS and then immediately back out and
		reduce to the LHS. We prefer to avoid pointless dancing about in the final automaton,
		so we detect the situation and correct for it by redirecting the RHS as if we were
		shifting the LHS instead, essentially running the reduction at table-generation time
		before a numbered state ever gets allocated.
		
		There are TWO little caveats:
		
		The target state, after traversing a unit rule, may or may not contain a reduction
		for that rule: this creates a slightly un-intuitive situation for the LALR construction.
		
		LR(1) and IE-LR(1) need to be able to find "iso-cores", so they must ALSO use this
		to avoid accidentally trying to find an iso-core that doesn't exist in the corresponding
		LR(0) automaton they use as an initial step.
		"""
		replace = {}
		for symbol in self.eligible_rhs & step.keys():
			each_item = iter(step[symbol])
			rule_id = next(each_item)[0]
			if rule_id in self.unit_rules and all(item[0] == rule_id for item in each_item):
				replace[symbol] = self.unit_rules[rule_id]
		shifts = {}
		for symbol in step.keys():
			proxy = symbol
			while proxy in replace: proxy = replace[proxy]
			shifts[symbol] = self.bft.lookup(frozenset(step[proxy]), breadcrumb=proxy)
		return shifts

class LA_State(NamedTuple):
	"""
	The key difference here from LR(0) is that the possible reductions are keyed
	to lookahead tokens from the follow-set of that reduction *however derived*.
	"""
	shift: Dict[str, int]
	reduce: Dict[str, List[int]]
	def reductions_before(self, lexeme): return self.reduce.get(lexeme, ())


def lalr_construction(grammar: context_free.ContextFreeGrammar) -> HFA[LA_State]:
	"""
	Building a non-deterministic LALR(1)-style table is a direct extension of the LR(0)
	construction. LR(0) tables tend to have lots of inadequate states. If we figure out
	which look-ahead tokens are relevant to which reductions, then the automaton gets a
	good deal more capable. In the limit, there is canonical LR(1) as given by Knuth.
	That's traditionally been considered impractical for all but the very smallest
	grammars. Today's workstations have the chops to handle canonical LR(1) even for
	larger grammars, but it's still a chore and, as we will see, a needless one.
	
	The bulk of the work involved consists of finding the follow-sets for reductions.
	I've chosen to break that work into its own function, rather than code the entire
	LALR construction as one function, for two reasons. First, it's an interesting
	algorithm in its own right. Second, that function provides the perfect level of
	abstraction for exploitation later in the minimal-LR(1) construction given later.
	
	If we stop here, we get a table that's more efficient than LR(0) for generalized
	parsing: it will attempt many fewer dead-end reductions before look-ahead tokens
	that LALR determines not to be in the reduction's follow-set.
	"""
	lr0 = lr0_construction(grammar)
	token_sets, follow = lalr_first_and_follow(lr0)
	def make_lalr_state(q:int, node:LR0_State) -> LA_State:
		reduce = {}
		for rule_id in node.reduce:
			for token in token_sets[follow[q, rule_id]]:
				assert isinstance(token, str)
				if token not in reduce: reduce[token] = [rule_id]
				else: reduce[token].append(rule_id) # This branch represents an R/R conflict...
		# At this point, `LA_State(node.shift, reduce)` would be a nice, potentially
		# non-deterministic HFA state incorporating information about LALR follow sets.
		# We can do better: Calling upon the "reachable" function causes the resulting
		# non-deterministic HFA to respect precedence and associativity (P&A) declarations.
		# This parallels the corresponding call in the true-LR(1) algorithm, simplifies
		# the construction of a deterministic table, does not change the deterministic
		# semantics, and may have some subtle implications for GLR semantics in the presence
		# of P&A declarations. In particular, an invasive high-precedence reduction could
		# cause trouble. I've not taken the time to work out if that's even a thing, though.
		# For now, if you want to do GLR parsing, use the minimal_lr1(...) construction.
		step = reachable(node.shift, reduce, grammar)
		# Having done that, it's entirely plausible that some unreachable states may
		# exist. It's probably not worth worrying about.
		return LA_State(step, reduce)
	return HFA(
		graph=[make_lalr_state(q, node) for q, node in enumerate(lr0.graph)],
		initial=lr0.initial, accept=lr0.accept, grammar=grammar, bft=lr0.bft
	)

def lalr_first_and_follow(lr0:HFA[LR0_State]) -> Tuple[list, dict]:
	"""
	This is a variant of the channel algorithm as described (colorfully and frankly not
	very clearly) in chapter 9 of Parsing Techniques: a Practical Guide, by Dick Grune
	and Ceriel Jacobs. In essence the idea is to build a directed graph by certain rules
	and then "flow" terminal symbols around the nodes in that graph until updates cease.
	You may recognize this as a fix-point over the set-union operation. It turns out we
	can do better: using Tarjan's Strongly-Connected-Components algorithm and orienting
	the edges in the correct direction, we need only ever consider each edge once.
	"""
	grammar = lr0.grammar
	terminals = grammar.apparent_terminals()
	token_sets = [terminals.intersection(node.shift.keys()) for node in lr0.graph]
	for q in lr0.accept:
		token_sets[q].add(interfaces.END_OF_TOKENS) # Denoting that end-of-input can follow a sentence.
	# Allocate and track/label all follow sets:
	follow = {}
	for q, node in enumerate(lr0.graph):
		for rule_id in node.reduce:
			follow[q,rule_id] = foundation.allocate(token_sets, set())
	# Generate a token-set inflow graph with nodes as lists of inbound edges.
	inflow:List[Set[int]] = [set() for _ in token_sets]
	for q, node in enumerate(lr0.graph):
		for symbol, target in node.shift.items():
			if symbol in grammar.symbol_rule_ids: # That is, if symbol is non-terminal,
				for rule_id in grammar.symbol_rule_ids[symbol]:
					rule_end = lr0.traverse(q, grammar.rules[rule_id].rhs)
					if (rule_end, rule_id) in follow: # Otherwise a unit-reduction was elided here.
						inflow[rule_end].add(target)
						inflow[follow[rule_end, rule_id]].add(target)
	# Determine the contents of all first- and follow-sets:
	for component in foundation.strongly_connected_components_by_tarjan(inflow):
		tokens = set()
		for k in component:
			tokens.update(token_sets[k])
			tokens.update(*[token_sets[j] for j in inflow[k]])
			token_sets[k] = tokens
	return token_sets, follow

def canonical_lr1(grammar: context_free.ContextFreeGrammar) -> HFA[LA_State]:
	"""
	Before embarking on a quest to produce a minimal-LR(1) table by sophisticated
	methods, it's worth learning how to produce the maximal-LR(1) table by (some
	variant of) Donald E. Knuth's original method.

	A Knuth parse-item is like an LR(0) item augmented with the (1) next token
	expected AFTER the corresponding rule would be recognized. The initial core
	would look like { .S/# } in the usual notation. Otherwise, the algorithm has
	much in common with the LR(0) construction above -- but to see that clearly
	you'll have to look at the function `abstract_lr1_construction(...)`.
	"""
	
	def front(symbol, follower, goto_transparent, iso_q):
		"""
		Return the list of parse-items which DIRECTLY follow from the given criteria.
		In Canonical LR(1), that means the beginnings of each rule for the
		prediction-symbol, with followers being every terminal that might come
		next after the prediction-symbol got shifted.
		"""
		iso_state = lr0.graph[iso_q]
		goto_state = lr0.graph[iso_state.shift[symbol]]
		after = set(goto_state.shift.keys()) & terminals
		if goto_transparent: after.add(follower)
		items = []
		for rule_id in grammar.symbol_rule_ids[symbol]:
			items.extend((rule_id, 0, lookahead) for lookahead in after)
		return items
	
	lr0 = lr0_construction(grammar) # This implicitly solves a lot of sub-problems.
	terminals = grammar.apparent_terminals()
	return abstract_lr1_construction(
		grammar, front = front,
		note_reduce = lambda reduce, follower, rule_id, iso_q: reduce[follower].append(rule_id),
		initial_item = lambda rule_id: (rule_id, 0, interfaces.END_OF_TOKENS),
		lr0_catalog=lr0.bft.catalog,
	)

def abstract_lr1_construction(
		grammar: context_free.ContextFreeGrammar, *,
		front, note_reduce, initial_item, lr0_catalog,
) -> HFA[LA_State]:
	"""
	The Canonical and Minimal LR(1) algorithms given here have a great deal in common.
	It seems both instructive and useful to factor out those commonalities.
	If you stare intently, you'll also see similarity to LR(0), but factoring
	out that particular commonality is not today's exercise.
	"""
	def build_state(core: frozenset):
		iso_q = lr0_catalog[frozenset((r, p) for (r, p, f) in core)]
		step, reduce = collections.defaultdict(set), collections.defaultdict(list)
		def visit(item):
			rule_id, position, follower = item
			rhs = RHS[rule_id]
			if position < len(rhs): # We have a non-terminal.
				next_symbol = rhs[position]
				step[next_symbol].add((rule_id, position+1, follower))
				if next_symbol in grammar.symbol_rule_ids:
					return front(next_symbol, follower, transparent(rule_id, position + 1), iso_q)
			elif rule_id<len(grammar.rules): note_reduce(reduce, follower, rule_id, iso_q)
		foundation.transitive_closure(core, visit)
		graph.append(LA_State(shift=ure.find_shifts(reachable(step, reduce, grammar)), reduce=reduce,))
	
	RHS = grammar.augmented_rules()
	transparent = find_transparent(grammar.find_first_and_epsilon()[1], RHS)
	bft = foundation.BreadthFirstTraversal()
	ure = UnitReductionEliminator(grammar, bft)
	graph = []
	initial = [bft.lookup(frozenset([initial_item(rule_id)])) for rule_id in grammar.initial()]
	bft.execute(build_state)
	accept = [graph[qi].shift[language] for qi, language in zip(initial, grammar.start)]
	# print("LR(0) states: %d\t\tLR(1) states:%d" % (len(lr0_catalog), len(graph)))
	return HFA(graph=graph, initial=initial, accept=accept, grammar=grammar, bft=bft)


def find_transparent(epsilon:Set[str], right_hand_sides:List[List[str]]):
	"""
	In LR(1)-family algorithms, we do a fair amount of tests for the rest-of-a-rule
	being transparent: i.e. all the remaining symbols in a right-hand-side (if any)
	being possibly-epsilon. The naive way -- something like
			all(symbol in epsilon for symbol in rhs[position:])
	works just fine, but think of the all those garbage list copies!
	We can do better. This is one way.
	"""
	def transparency_threshold(rhs):
		position = len(rhs)
		# Invariant: all(symbol in epsilon for symbol in rhs[position:])
		while position > 0 and rhs[position-1] in epsilon: position -= 1
		return position
	thresholds = list(map(transparency_threshold, right_hand_sides))
	return lambda rule_id, position: thresholds[rule_id] <= position


def minimal_lr1(grammar: context_free.ContextFreeGrammar) -> HFA[LA_State]:
	"""
	This amounts to a hybrid of LALR and Canonical LR(1) in which only the conflicted
	parts are reconsidered in greater detail. Details of the approach are in the
	doc-comments for the `front` and `note_reduce` subroutines.
	
	After poring over the IELR paper on several occasions, I believe there may yet be
	some originality in this contribution. If anything, I have an argument why this
	routine produces absolutely minimal tables: each output state has an absolutely
	minimal set of "siblings" affiliated with a corresponding LALR state, because the
	only thing distinguishing "siblings" is the correct final deterministic parse
	action after the rule associated with a parse-item is recognized, and then only
	for those (generally very few) tokens for which LALR does not figure it out. Also,
	unit rule elimination is applied, and unreachable states (in light of conflict
	resolutions) are never considered.
	"""
	def front(symbol, follower, goto_transparent, iso_q):
		"""
		Return the list of parse-items which DIRECTLY follow from the given criteria.
		
		The complexity here comes from how this algorithm threads the needle between
		LALR-when-adequate and LR(1)-when-necessary.
		"""
		isostate = lr0.graph[iso_q]
		items = []
		goto_q = isostate.shift[symbol]
		goto_shifts = lr0.graph[goto_q].shift.keys()
		goto_conflict = conflict_data[goto_q].tokens.keys()
		for sub_rule_id in grammar.symbol_rule_ids[symbol]:
			# Most of the smarts in this algorithm comes down to understanding what
			# LALR found at the far end of each sub-production. We need to know which
			# LR(0) state you reach after shifting the contents of that sub-rule:
			reach = lr0.traverse(iso_q, grammar.rules[sub_rule_id].rhs)
			if follower is None:  # We're coming from LALR-land:
				items.append((sub_rule_id, 0, None))
				reach_conflict = conflict_data[reach].rules.get(sub_rule_id, EMPTY)
				for token in reach_conflict & goto_shifts - goto_conflict:
					# NB: It's tempting to exclude candidate follower-tokens for which
					# a shift is assured according to the precedence declarations, but
					# it doesn't buy you anything (S/R conflicts get resolved later)
					# and in practice it results in larger parse tables.
					# We leave out the "GOTO-conflicted" tokens here because...
					items.append((sub_rule_id, 0, token)) # Canonical parse-item
			else:  # The canonical branch:
				# GOTO-conflicted tokens will have resulted in canonical-style parse items.
				# As with Canonical, they can follow a derivation only when the remainder
				# of the current rule is "transparent", but this algorithm imposes the
				# additional constraint regarding the token's contribution to a
				# LALR-inadequacy in the "reach" state.
				if follower in conflict_data[reach].tokens and goto_transparent:
					assert follower in goto_conflict
					items.append((sub_rule_id, 0, follower))
		return items
	
	def note_reduce(reduce, follower, rule_id, iso_q):
		"""
		There are two cases:
		
		If the "follower" is `None`, it stands for the un-conflicted portion of the
		corresponding LALR follow set.
		
		Otherwise, the token MUST have earlier been implicated in a LALR-inadequacy
		in this state (which fact we assert for good measure). Handle it the same as
		Canonical-LR(1).
		
		Incidentally, it is possible to reach a particular `reduce[follower]` list
		more than once if and only if the follower is LALR-inadequate. Proof follows
		from the fact that a given parse-item is visited at most once.
		"""
		if follower is None:
			for t in token_sets[follow[iso_q, rule_id]] - conflict_data[iso_q].rules[rule_id]:
				assert t not in reduce
				reduce[t] = [rule_id]
		else:
			assert follower in conflict_data[iso_q].rules[rule_id]
			if follower in reduce:
				assert rule_id not in reduce[follower]
				reduce[follower].append(rule_id)
			else: reduce[follower] = [rule_id]
	
	EMPTY = frozenset()
	lr0 = lr0_construction(grammar) # This implicitly solves a lot of sub-problems.
	token_sets, follow = lalr_first_and_follow(lr0)
	# Later we need to know if a certain rule is implicated in an LALR conflict: if so, for which terminals?
	# We also need to know if a state is conflicted with respect to a particular terminal.
	conflict_data = find_conflicts(lr0.graph, {(q,r):token_sets[i] for (q,r),i in follow.items()}, grammar)

	return abstract_lr1_construction(
		grammar,
		front = front, note_reduce = note_reduce,
		initial_item = lambda rule_id: (rule_id, 0, None),
		lr0_catalog=lr0.bft.catalog,
	)

def reachable(step:dict, reduce:dict, grammar) -> dict:
	"""
	The object of this function is to prevent the exploration of useless/unreachable states.
	It does this by deleting a shift from the "step" dictionary whenever said shift becomes
	impossible by virtue of P&A declarations. It also deletes useless reductions similarly
	rendered unavailable, and denotes non-associativity errors by a sentinel empty-tuple.
	
	This was developed to help minimal-LR1 mode respect precedence/associativity, but it's
	equally useful for all modes.
	
	NOTE: When shift/reduce/reduce conflicts arise in connection with operator-precedence
	specifications, the intended semantics for a generalized parse are not always clearly
	defined. The code below should either behave sensibly or toss an exception.
	
	The problem with the bizarre corner cases is they cannot be understood solely in terms
	of actions the parser might (or might not) take in this state. If they come up, you
	get a warning printed on STDERR but otherwise the operator-precedence declarations
	are ignored in that instance.
	"""
	for token, rule_id_list in list(reduce.items()):
		if token not in step: continue
		decide = [grammar.decide_shift_reduce(token, rule_id) for rule_id in rule_id_list]
		ways = set(decide)
		assert context_free.BOGUS not in ways, "This is guaranteed by grammar.validate(...), called earlier."
		if len(ways) == 1:
			decision = ways.pop()
			if decision == context_free.LEFT: del step[token]
			elif decision == context_free.RIGHT: del reduce[token]
			elif decision == context_free.NONASSOC:
				del step[token]
				reduce[token] = ()
			else: assert decision is None
		elif ways == {context_free.LEFT, context_free.NONASSOC}:
			del step[token]
			reduce[token] = tuple(r for r,d in zip(rule_id_list, decide) if d == context_free.LEFT)
		elif ways == {context_free.RIGHT, None}:
			reduce[token] = tuple(r for r, d in zip(rule_id_list, decide) if d != context_free.RIGHT)
		else:
			print("Fair Warning:", token, "triggers a bizarre operator-precedence corner case.", file=sys.stderr)
	return step


class ConflictData(NamedTuple):
	tokens: Dict[str, Set[int]] # The rules that conflict on this token
	rules: Dict[int, Set[str]] # The tokens that conflict on this rule.

def find_conflicts(graph, follow_sets, grammar) -> List[ConflictData]:
	"""
	This drives one of the central ideas of the Minimal-LR(1) algorithm:
	Learn which tokens are involved in conflicts, and which rules contribute
	to those conflicts for each token (as known to LALR).
	
	Subtleties:
	
	1. If an S/R conflict is DECLARED to shift, then it does not impugn the token,
	   but the token still refers to the rule in case some other rule MAY reduce.
	
	This routine (and the stuff that uses it) is coded with the idea that the
	grammar may leave certain things deliberately non-deterministic. Wherever that
	is the case, these algorithms will respect it.
	
	2. There is a way to improve the treatment of R/R conflicts if it is known in
	   advance that the table will be used deterministically, or if the R/R is
	   resolved by rule precedence. It involves a pass over the LR(1) item cores
	   considering the groups that eventually lead to a R/R conflict (they have the
	   same suffix and follower): among those groups only the "winning" reduction
	   item needs to stay in the core. This "normalizes" the LR(1) cores so that
	   potentially fewer distinct ones might be generated.
	
	Alas, idea #2 is not yet implemented. Complex R/R conflicts may still lead to
	more states than strictly necessary for a deterministic table. In practice,
	this is unlikely to be a real problem: deterministic tables are usually made
	from grammars with few LALR conflicts, and in the non-deterministic case
	nothing is wasted. Nevertheless, this remains an avenue for improvement.
	"""
	result = []
	for q, state in enumerate(graph):
		degree = collections.Counter(state.shift.keys()) # This picks up some nonterminals but they do no harm.
		for rule_id in state.reduce:
			for token in follow_sets[q,rule_id]:
				prefer_shift = token in state.shift and grammar.decide_shift_reduce(token, rule_id) == context_free.RIGHT
				if not prefer_shift: degree[token] += 1
		conflicted_tokens = set(token for token, count in degree.items() if count > 1)
		conflict = ConflictData({token: set() for token in conflicted_tokens}, {})
		for rule_id in state.reduce:
			contribution = conflicted_tokens & follow_sets[q,rule_id]
			conflict.rules[rule_id] = contribution
			for token in contribution: conflict.tokens[token].add(rule_id)
		result.append(conflict)
	return result


class DragonBookTable(interfaces.ParseTable):
	"""
	This is the classic textbook view of a set of parse tables: a pair of dense matrices
	(implemented here as lists-of-lists) representing the "ACTION" and "GOTO" tables, along
	with information about the reduction rules. The contents of these matrices are just
	numbers representing parse actions.
	
	This is a reasonable implementation as-is if you have a modern amount of RAM in your
	machine. In days of old, it would be necessary to compress the parse tables. Today,
	that's still not such a bad idea if you can pre-compute the tables. The compaction
	submodule contains some code for a typical method of parser table compression, and
	the runtime submodule implements the ParseTable interface atop a compressed table.
	
	Design note: There's a temptation to make the constructor take an HFA object, but
	that limits the ways you can instantiate this class. See the function `tabulate(...)`.
	"""
	
	def __init__(self, *, initial: dict, action: list, goto: list, nonassoc_errors: set, rules: list, terminals: list,
	             nonterminals: list, breadcrumbs: list, splits=()):
		self.initial = initial
		self.action_matrix = action
		self.goto_matrix = goto
		self.nonassoc_errors = nonassoc_errors
		self.translate = {symbol: i for i, symbol in enumerate(terminals)}
		nontranslate = {symbol: i for i, symbol in enumerate(nonterminals)}
		self.terminals, self.nonterminals = terminals, nonterminals
		self.breadcrumbs = breadcrumbs
		self.constructors = [None, *(set(rule.constructor for rule in rules) - {None})]
		ctrans = {c:i for i,c in enumerate(self.constructors)}
		def translate_rule(rule:context_free.Rule):
			size = len(rule.rhs)
			if isinstance(rule.places, int):
				assert rule.constructor is None
				c,p = rule.places-size, ()
			else: c,p = ctrans[rule.constructor], tuple(x-size for x in rule.places)
			return nontranslate[rule.lhs], size, c, p
		self.rule_table = list(map(translate_rule, rules))
		self.rule_origin = [rule.origin for rule in rules]
		self.splits = splits # A non-deterministic table just needs one extra bit: this list of lists.
		
		interactive = []
		for row in action:
			k = set(row)
			k.discard(0)
			if len(k) == 1: interactive.append(min(k.pop(), 0))
			else: interactive.append(0)
		for q, t in nonassoc_errors: interactive[q] = False
		self.interactive_step = self.interactive_rule_for = interactive.__getitem__
	
	def get_rule(self, rule_id: int) -> tuple:
		return self.rule_table[rule_id]
	
	def get_translation(self, symbol) -> int:
		try: return self.translate[symbol]
		except KeyError: return len(self.terminals) # Guaranteed to trigger error-processing.
		
	def get_action(self, state_id, terminal_id) -> int: return self.action_matrix[state_id][terminal_id]
	
	def get_goto(self, state_id, nonterminal_id) -> int: return self.goto_matrix[state_id][nonterminal_id]
	
	def get_initial(self, language) -> int: return 0 if language is None else self.initial[language]
	
	def get_breadcrumb(self, state_id) -> str: return self.breadcrumbs[state_id]
	
	def display(self):
		size = len(self.action_matrix)
		print('Action and Goto: (%d states)' % size)
		head = ['', ''] + self.terminals + [''] + self.nonterminals
		body = []
		for i, (b, a, g) in enumerate(zip(self.breadcrumbs, self.action_matrix, self.goto_matrix)):
			body.append([i, b, *a, '', *g])
		pretty.print_grid([head] + body)
	
	def make_csv(self, pathstem):
		""" Generate action and goto tables into CSV files suitable for inspection in a spreadsheet program. """
		
		def mask(q, row, essential):
			return [
				s if s or (q, t) in essential else None
				for t, s in enumerate(row)
			]
		
		def typical_grid(top, matrix, essential):
			head = [None, None, *top]
			return [head] + [[q, self.breadcrumbs[q]] + mask(q, row, essential) for q, row in enumerate(matrix)]
		
		pretty.write_csv_grid(pathstem + '.action.csv',
			typical_grid(self.terminals, self.action_matrix, self.nonassoc_errors))
		pretty.write_csv_grid(pathstem + '.goto.csv', typical_grid(self.nonterminals, self.goto_matrix, frozenset()))
	
	def get_split_offset(self) -> int:
		return len(self.action_matrix)
	
	def get_split(self, split_id: int) -> list:
		assert split_id>0
		return self.splits[split_id]




class ParsingStyle:
	"""
	There are three main ways to deal with inadequacies (non-determinism) remaining
	after application of any P&A declarations:
		1. Pure: Inadequacies are considered a grammar bug.
		2. Deterministic: Inadequacies are resolved to shift, or to use the earliest-defined rule.
		3. Generalized: Inadequacies are converted to parser-split entries.
	
	Probably the correct choice of style should be reflected in the grammar definition somehow.
	"""
	
	def decide_inadequacy(self, q:int, look_ahead:str, shift:int, rule_ids:Iterable, rules:list) -> int:
		""" Called in all non-deterministic situations. """
		raise NotImplementedError(type(self))
	
	def any_splits(self):
		""" Return nothing, or a list of splits for use in non-deterministic parsing algorithms. """
		raise NotImplementedError(type(self))

	def report(self, hfa):
		""" Give user-feedback about any observed challenges. """
		raise NotImplementedError(type(self))

class DeterministicStyle(ParsingStyle):
	
	def __init__(self, strict:bool):
		self.conflicts = collections.defaultdict(list)
		self.strict = strict
	
	def decide_inadequacy(self, q:int, look_ahead: str, shift: int, rule_ids: Iterable, rules:list) -> int:
		self.conflicts[q, look_ahead].extend(rule_ids)
		return shift or encode_reduce(min(rule_ids))
	
	def any_splits(self):
		pass
	
	def report(self, hfa):
		"""
		This function was originally intended as a way to visualize the branches of a conflict.
		In its original form a bunch of context was available; I've gratuitously stripped that away
		and now I want to break this down to the bits we actually need.
		
		BreadthFirstTraversal.traversal[x] was used to grab the core parse items in order to
		visualize the state reached by shifting the lookahead token if that shift is viable.
		Such really belongs as a method on the state: soon it will move there.
		
		The "options" list contains numeric candidate ACTION instructions which are interpreted
		in the usual way: This does represent a data-coupling, but one that's unlikely to change,
		so I'm not too worried about it just now.
		
		In conclusion: Let the objects defined in automata.py format parse-states for human consumption.
		"""
		for (q, lookahead), rule_ids in sorted(self.conflicts.items()):
			hfa.display_situation(q, lookahead)
			if lookahead in hfa.graph[q]:
				shift = hfa.graph[q][lookahead]
				print("Do we shift into:")
				left_parts, right_parts = [], []
				for r, p in hfa.earley_core(shift):
					rhs = hfa.grammar.rules[r].rhs
					left_parts.append(' '.join(rhs[:p]))
					right_parts.append(' '.join(rhs[p:]))
				align = max(map(len, left_parts)) + 10
				for l, r in zip(left_parts, right_parts):
					print(' ' * (align - len(l)) + l + '  ' + pretty.DOT + '  ' + r)
			for r in rule_ids:
				rule = hfa.grammar.rules[r]
				print("Do we reduce:  %s -> %s" % (rule.lhs, ' '.join(rule.rhs)))
		if self.strict and self.conflicts: raise PurityError()


class GeneralizedStyle(ParsingStyle):
	
	def __init__(self, splits_offset:int, nondeterministic_symbols:set):
		self.offset = splits_offset
		self.splits = []
		self.nondeterministic_symbols = nondeterministic_symbols
	
	def decide_inadequacy(self, q: int, look_ahead: str, shift: int, rule_ids: Iterable, rules:list) -> int:
		split = []
		if shift: split.append(shift)
		for r in sorted(rule_ids, key=lambda i:len(rules[i].rhs)): split.append(-1-r)
		return self.offset + foundation.allocate(self.splits, split)
	
	def any_splits(self):
		return self.splits
	
	def report(self, hfa):
		print(len(self.splits), "non-deterministic situation(s) encountered.")


def encode_reduce(rule_id:int) -> int:
	""" See interface.ParseTable.get_action. """
	return -1 - rule_id

def tabulate(hfa: HFA[LA_State], *, style:ParsingStyle) -> DragonBookTable:
	"""
	Having an HFA based on State objects, this function produces a corresponding
	dense-matrix-style parse table of the sort typically shown in textbook descriptions
	of LR-style parsing automata.
	
	This function does NOT worry about precedence and associativity declarations:
	It assumes that concern has already been taken care of in the input HFA -
	principally by the interaction of function `reachable(...)` with the P&A bits.
	
	Any residual inadequacies of the grammar are delegated to the `style` object for
	resolution.
	"""
	grammar = hfa.grammar
	assert interfaces.END_OF_TOKENS not in grammar.symbols
	assert interfaces.ERROR_SYMBOL not in grammar.symbol_rule_ids
	terminals = [interfaces.END_OF_TOKENS] + sorted(grammar.apparent_terminals())
	translate = {t:i for i,t in enumerate(terminals)}
	nonterminals = sorted(grammar.symbol_rule_ids.keys())
	##### Tabulate the states into dense matrices ACTION and GOTO:
	action, goto, nonassoc_errors = [], [], set()
	conflict = collections.defaultdict(set)
	for q, state in enumerate(hfa.graph):
		goto.append([state.shift.get(s, 0) for s in nonterminals])
		action_row = [state.shift.get(s, 0) for s in terminals]
		conflict.clear()
		for symbol, rule_ids in state.reduce.items():
			idx = translate[symbol]
			shift = action_row[idx]
			if rule_ids == ():
				# This is how function `reachable(...)` communicates a non-association situation.
				assert shift == 0
				nonassoc_errors.add((q,idx))
			elif shift == 0 and len(rule_ids) == 1:
				action_row[idx] = encode_reduce(rule_ids[0])
			else: action_row[idx] = style.decide_inadequacy(q, symbol, shift, rule_ids, grammar.rules)
		action.append(action_row)
	for q, t in nonassoc_errors: action[q][t] = 0
	for q in hfa.accept: action[q][0] = q
	style.report(hfa)
	return DragonBookTable(
		initial=dict(zip(grammar.start, hfa.initial)),
		action=action,
		goto=goto,
		nonassoc_errors=nonassoc_errors,
		rules=grammar.rules,
		terminals=terminals,
		nonterminals=nonterminals,
		breadcrumbs=hfa.bft.breadcrumbs,
		splits=style.any_splits()
	)


PARSE_TABLE_METHODS = {
	'LALR': lalr_construction,
	'CLR': canonical_lr1,
	'LR1': minimal_lr1,
}





"""
Attributed context-free grammar with operator-precedence and associativity declarations.

This file is basically all about the data structure, with relatively little intrinsic
functionality beyond simple queries. That's because the kinds of work people might do
to a model of a grammar extends far beyond the specific applications here included.
"""

import collections, itertools
from typing import List, Optional, NamedTuple, Hashable, Sequence
from ..support import foundation, pretty

LEFT, RIGHT, NONASSOC, BOGUS = object(), object(), object(), object()

class Fault(ValueError): pass
class DuplicateRule(Fault): pass
class NonTerminalsCannotHavePrecedence(Fault): pass
class FailedToSetPrecedenceOnPrecsym(Fault): pass
class PrecedenceDeclaredTwice(Fault): pass
class RuleProducesBogusToken(Fault): pass # The rule produces a bogus symbol...
class UnreachableSymbols(Fault): pass
class IllFoundedSymbols(Fault): pass
class RenamingLoop(Fault): pass
class EpsilonLoop(Fault): pass

class ContextFreeGrammar:
	"""
	Formally, a context free grammar consists of:
		* A set of terminal symbols,
		* A set of non-terminal symbols, disjoint from the terminals,
		* A set of production rules (as described at class Rule),
		* and a start symbol.
	
	As a cool and potentially useful hack, I expand the normal definition to allow
	potentially several "start symbols" to share the same overall definition. This allows
	a single set of tables to be used (with a different initial state) to parse different
	languages (or sub-phrases) while sharing common specification elements.
	
	Practical languages usually have some ambiguities which are ordinarily resolved with
	precedence declarations. If explicit declarations are not provided, the grammar is
	ambiguous -- at least with respect to the power of the table-generation method used.
	There are ways to deal with ambiguity, but for deterministic parsers the convention
	is to shift on a shift-reduce conflict (i.e. assume right-associativity) and, in
	the event of a reduce/reduce conflict, to use the earliest-defined rule.
	"""
	def __init__(self):
		self.symbols = set()
		self.start = [] # The start symbol(s) may be specified asynchronously to construction or the rules.
		self.rules, self.token_precedence, self.level_assoc = [], {}, []
		self.symbol_rule_ids = {}
	def display(self):
		head = ['', 'Symbol', 'Produces', 'Using']
		body = [[i, rule.lhs, rule.rhs, rule.attribute] for i,rule in enumerate(self.rules)]
		pretty.print_grid([head] + body)

	def rule(self, lhs:str, rhs:List[str], prec_sym, constructor:Hashable, places:(list, tuple, int), origin):
		"""
		This is your basic mechanism to install arbitrary plain-jane BNF rules.
		It provides the consistency checks better than if you were to instantiate Rule directly.
		
		For the sake of simplicity, at this layer symbols are all strings.
		Duplicate rules are rejected by raising an exception.
		
		:param lhs: The "left-hand-side" non-terminal symbol which is declared to produce...
		
		:param rhs: this "right-hand-side" sequence of symbols.
		
		:param prec_sym: Set the precedence of this reduction explicitly according to that
		of the given symbol as previously declared. If this is not supplied and the table
		construction algorithm finds it necessary, method infer_prec_sym(...) implements
		default processing on the right-hand-side.
		
		:param constructor: and
		:param places: designate the semantics for the attribute synthesis portion of the
		rule specification. If `constructor` is `None`, then places is expected to be the
		significant-symbol index into the RHS for a renaming or bracketing rule. Otherwise,
		`places` should be a list/tuple of such indices relevant to the given constructor.
		Note: At this level we don't interpret constructors further.
		
		Note: unit/renaming productions are recognized, and the parser tables will optimize
		such rules to a zero-cost abstraction wherever possible.
		
		:param origin: Use this to indicate the provenance of the rule. It can be a source
		line number, or whatever else makes sense in your application.
		
		:return: Nothing.
		
		Please note: Certain rare constructions make the unit-rule optimization unsound.
		The tables will be constructed correctly but the general parsing algorithm needs
		to be prepared to find a rule with a null attribute. Correct behavior is generally
		to leave the semantic-stack unchanged.
		"""
		if lhs in self.token_precedence: raise NonTerminalsCannotHavePrecedence(lhs)
		self.symbols.add(lhs)
		self.symbols.update(rhs)
		if lhs not in self.symbol_rule_ids: self.symbol_rule_ids[lhs] = []
		if isinstance(places, int): assert constructor is None and 0 <= places < len(rhs)
		else: assert isinstance(places, (list, tuple)) and all(p<len(rhs) for p in places), places
		sri = self.symbol_rule_ids[lhs]
		if any(self.rules[rule_id].rhs == rhs for rule_id in sri): raise DuplicateRule(lhs, rhs)
		sri.append(foundation.allocate(self.rules, Rule(lhs, rhs, prec_sym, constructor, places, origin)))
	
	def augmented_rules(self) -> list:
		"""
		The LR family of algorithms are normally explained in terms of an augmented grammar:
		It has a special "accept" rule which just expands to the start symbol for the ordinary
		context-free grammar at issue. There's a good reason for this: It makes the algorithms
		work properly without a lot of special edge cases to worry about. However, the "accept"
		non-terminal ought to be imaginary: it can't appear on the right, one never reduces the
		"accept" symbol, and I object to it consuming space in the GOTO table as it typically
		does in implementations that rely on a first-class rule (with left-hand side symbol).

		This object exists to clarify interaction with the "augmented" rule set.
		NB: Since we support multiple "start" symbols, there are corresponding "accept" rules.
		"""
		assert self.start
		return [rule.rhs for rule in self.rules] + [[language] for language in self.start]
	
	def initial(self) -> range:
		""" The range of augmented-rule indices corresponding to the list of start symbols. """
		first = len(self.rules)
		return range(first, first+len(self.start))
	
	def apparent_terminals(self) -> set:
		""" Of all symbols mentioned, those without production rules are apparently terminal. """
		return self.symbols - self.symbol_rule_ids.keys()

	def find_first_and_epsilon(self):
		"""
		Answers the two questions:
			Which terminals can symbol X begin with?
			Which symbols can produce the empty string?
		This solution takes pains not to repeat work, and so should be reasonably quick.
		"""
		epsilon = set()
		first = {s:{s} for s in self.symbols}
		hangar = collections.defaultdict(list)
		work = [(i,0) for i in range(len(self.rules))]
		while work:
			r,p = work.pop()
			lhs, rhs, _, _, _, _ = self.rules[r]
			if p == len(rhs):
				epsilon.add(lhs)
				if lhs in hangar: work.extend(hangar.pop(lhs))
			else:
				symbol = rhs[p]
				first[lhs].add(symbol)
				if symbol in epsilon: work.append((r,p+1))
				else: hangar[symbol].append((r,p+1))
		for component in foundation.strongly_connected_components_hashable(first):
			f = set()
			for symbol in component:
				f.update(*(first[x] for x in first[symbol]))
				first[symbol] = f
			f.difference_update(self.symbol_rule_ids)
		return first, epsilon
		
	def assert_no_bogons(self):
		""" "Bogus" tokens only exist to establish precedence levels and must not appear in right-hand sides. """
		bogons = {sym for sym, prec in self.token_precedence.items() if self.level_assoc[prec] is BOGUS}
		for rule_id, rule in enumerate(self.rules):
			if any(symbol in bogons for symbol in rule.rhs):
				raise RuleProducesBogusToken(rule_id)
	
	def assert_well_founded(self):
		"""
		Here, "well-founded" means "can possibly produce a finite sequence of terminals."
		"Ill-founded" is the opposite. Example ill-founded grammars:
			S -> x S
			A -> B y;  B -> A x
		
		A terminal symbol is well-founded.
		A rule with only well-founded symbols in the right-hand side is well-founded.
		A non-terminal symbol with at least one well-founded rule is well-founded.
		Induction applies. A grammar with only well-founded symbols is well-founded.
		"""
		well_founded = self.apparent_terminals()
		black = list(self.rules)
		while black:
			red = []
			for rule in black:
				if rule.lhs not in well_founded:
					if all(s in well_founded for s in rule.rhs): well_founded.add(rule.lhs)
					else: red.append(rule)
			if len(red) < len(black): black = red
			else: raise IllFoundedSymbols(set(rule.lhs for rule in red))
	
	def assert_no_orphans(self):
		""" Every symbol should be reachable from the start symbol(s). """
		produces = collections.defaultdict(set)
		for rule in self.rules: produces[rule.lhs].update(rule.rhs)
		unreachable = self.symbols - foundation.transitive_closure(self.start, produces.get)
		if unreachable: raise UnreachableSymbols(unreachable) # NB: Bogons are not among self.symbols.
	
	def assert_no_rename_loops(self):
		""" If a symbol may be replaced by itself (possibly indirectly) then it is diseased. """
		broken = set()
		renames = collections.defaultdict(set)
		for lhs, rhs, _, _, _, _ in self.rules:
			if len(rhs) == 1:
				if lhs == rhs[0]: broken.add(lhs)
				else: renames[lhs].add(rhs[0])
		for component in foundation.strongly_connected_components_hashable(renames):
			if len(component) > 1: broken.update(component)
		if broken: raise RenamingLoop(broken)
	
	def assert_no_epsilon_loops(self):
		""" Epsilon Left-Self-Recursion is OK. All other recursive-epsilon-loops are pathological. """
		first, epsilon = self.find_first_and_epsilon()
		reaches = collections.defaultdict(set)
		broken = set()
		for lhs, rhs, _, _, _, _ in self.rules:
			epsilon_prefix = list(itertools.takewhile(epsilon.__contains__, rhs))
			if not epsilon_prefix: continue
			if epsilon_prefix[0] == lhs: epsilon_prefix.pop(0)
			if lhs in epsilon_prefix: broken.add(lhs)
			reaches[lhs].update(epsilon_prefix)
		for component in foundation.strongly_connected_components_hashable(reaches):
			if len(component) > 1: broken.update(component)
		if broken: raise EpsilonLoop(broken)
		
	def validate(self):
		"""
		This raises an exception (derived from Fault, above) for the first error noticed.
		It might be nice to get a more complete read-out of problems, but that would mean
		inventing some sort of document structure to talk about faults in grammars.
		Then again, maybe that's in the pipeline.
		"""
		self.assert_no_bogons()
		self.assert_well_founded()
		self.assert_no_orphans()
		self.assert_no_rename_loops()
		self.assert_no_epsilon_loops()
	
	def assoc(self, direction, symbols):
		assert direction in (LEFT, NONASSOC, RIGHT, BOGUS)
		assert symbols
		level = foundation.allocate(self.level_assoc, direction)
		for symbol in symbols:
			if symbol in self.symbol_rule_ids: raise NonTerminalsCannotHavePrecedence(symbol)
			if symbol in self.token_precedence: raise PrecedenceDeclaredTwice(symbol)
			self.token_precedence[symbol] = level
			
	def decide_shift_reduce(self, symbol, rule_id):
		try: sp = self.token_precedence[symbol]
		except KeyError: return None
		rp = self.determine_rule_precedence(rule_id)
		if rp is None: return None
		if rp < sp: return LEFT
		# NB: Bison and Lemon both treat later declarations as higher-precedence,
		# which is unintuitive, in that you perform higher-precedence operations
		# first so it makes sense to list them first. Please excuse my dear aunt Sally!
		if rp == sp: return self.level_assoc[rp]
		return RIGHT
	
	def infer_prec_sym(self, rhs):
		"""
		If a rule without an explicit precedence declaration is involved in a shift/reduce conflict,
		the parse table generation algorithm will call this to decide which symbol represents the
		precedence of this right-hand-side.
		
		As a slight refinement to the BISON approach, this returns the first terminal with an
		assigned precedence, as opposed to the first terminal symbol whatsoever. Often that's
		a distinction without a difference, but when it matters I think this makes more sense.
		"""
		for symbol in rhs:
			if symbol in self.token_precedence:
				return symbol
	
	def determine_rule_precedence(self, rule_id):
		rule = self.rules[rule_id]
		prec_sym = rule.prec_sym or self.infer_prec_sym(rule.rhs)
		if prec_sym: return self.token_precedence[prec_sym]

class Rule(NamedTuple):
	# The first few fields define the actual grammar
	lhs: str
	rhs: List[str]
	prec_sym: Optional[str]
	# The next couple deal with the idea of attribute synthesis
	constructor: Hashable
	places: Sequence[int]
	# Last is all about the provenance of the rule, which is important in debugging.
	origin: object
	
	def is_rename(self):
		return self.constructor is None and len(self.rhs) == 1 and self.places == 0




"""
This module implements the essential algorithm to recognize a sequence of
lexemes and their syntactic categories (rules/scan-actions) by reference to a
finite-state machine. It supports backtracking, beginning-of-line anchors,
and (non-variable) trailing context.
"""

from ..support import interfaces

END_OF_INPUT = -1 # Used in place of a character ordinal. Agrees with the DFA builder.

class IterableScanner(interfaces.Scanner):
	"""
	This is the standard generic finite-automaton-based scanner, with support for backtracking,
	beginning-of-line anchors, and (non-variable) trailing context.
	
	Your application must provide a suitable finite-automaton, rule bindings, and error handler.
	"""
	def __init__(self, *, text:str, automaton: interfaces.FiniteAutomaton, rules: interfaces.ScanRules, start, on_error:interfaces.ScanErrorListener):
		if not isinstance(text, str): raise ValueError('text argument should be a string, not a ', type(text))
		self.__text = text
		self.__automaton = automaton
		self.__rules = rules
		self.on_error = on_error
		self.enter(start)
		self.__stack = []
		self.__start, self.__mark = None, None
		self.__buffer = []
	
	def enter(self, condition):
		self.__condition_name = condition
		self.__condition = self.__automaton.get_condition(condition)
		
	def pop(self):
		self.enter(self.__stack.pop())
		
	def push(self, condition):
		self.__stack.append(self.__condition_name)
		self.enter(condition)
		
	def current_condition(self):
		return self.__condition_name
	
	def matched_text(self) -> str:
		""" As advertised. """
		return self.__text[self.__start:self.__mark]
	
	def less(self, nr_chars:int):
		"""
		Put trailing characters back into the stream to be matched.
		This is also the mechanism for trailing context.
		if `nr_chars` is zero or positive, the current match is adjusted to
		consume precisely that many characters. If it's negative, then a
		corresponding number of characters are dropped from the end of the
		match, and these will be considered again in the next matching cycle.
		"""
		new_mark = (self.__mark if nr_chars < 0 else self.__start) + nr_chars
		assert self.__start <= new_mark <= self.__mark
		self.__mark = new_mark
	
	def scan_one_raw_lexeme(self, q) -> int:
		"""
		This is how the magic happens. Given a starting state, this algorithm finds the
		end of the lexeme to be matched and the rule number (if any) that applies to the match.
		
		This algorithm deliberately ignores a zero-width, zero-context match, but it may fail
		to consume any characters (returning None). The caller must deal properly with that case.
		"""
		text, automaton = self.__text, self.__automaton
		cursor, mark, rule_id, jam = self.__start, self.__start, None, automaton.jam_state()
		while True:
			try: codepoint = ord(text[cursor])
			except IndexError: codepoint = END_OF_INPUT # EAFP: Python will check string length anyway...
			q = automaton.get_next_state(q, codepoint)
			if q == jam: break
			cursor += 1
			q_rule = automaton.get_state_rule_id(q)
			if q_rule is not None: mark, rule_id = cursor, q_rule
		self.__mark = mark
		return rule_id
	
	def __q0(self):
		""" Return the current "initial" state of the automaton, based on condition and context. """
		at_begin_line = self.__start == 0 or self.__text[self.__start - 1] in '\r\n'
		return self.__condition[at_begin_line]
	
	def token(self, kind, semantic=None):
		""" Be it established, then, that the token stream shall consist of... """
		assert kind is not None
		self.__buffer.append((kind, semantic))
	
	def __iter__(self):
		"""
		It seems convenient that iterating over the scanner should cause it to scan...
		"""
		
		def fire_rule(rule_id):
			"""
			The process to fire a rule is somewhat distinct from the decision.
			There are a couple bits
			"""
			if rule_id is None:
				self.__mark = self.__start + 1 # Prepare to "match" the offending character...
				self.on_error.scan_blocked(self) # and delegate to the error handler, which may do anything.
			else:
				trail = rules.get_trailing_context(rule_id)
				if trail is not None: self.less(trail)
				try: rules.invoke(self, rule_id)
				except Exception as ex: self.on_error.scan_exception(self, rule_id, ex)
				yield from self.__buffer
				self.__buffer.clear()

		automaton, rules = self.__automaton, self.__rules
		self.__start, eot = 0, len(self.__text)
		while self.__start < eot:
			yield from fire_rule(self.scan_one_raw_lexeme(self.__q0()))
			self.__start = self.__mark
		# Now determine if an end-of-file rule needs to execute:
		q = automaton.get_next_state(self.__q0(), END_OF_INPUT)
		if q>=0: yield from fire_rule(automaton.get_state_rule_id(q))
		
	def current_position(self) -> int:
		""" As advertised. This was motivated by a desire to produce helpful error messages. """
		return self.__start
	def current_span(self):
		""" Return the position and length of the current match-text for use in error-reporting calls and the like. """
		return self.__start, self.__mark - self.__start


"""
This file gives the essential shift-reduce algorithm for parsing with with LR-family parse tables.
The exact form of the table is not important to this exposition, so this module treats the parse
table as an abstract data type defined and implemented elsewhere.

A word on reductions:
-------------------------
There must be a table of rules consisting of:
	left-hand side non-terminal symbol (an ID number),
	rule length (the number of states to pop after computing the reduction),
	"message" -- generally, a "constructor" symbol and a vector of stack offsets to parameters.

As a minor optimization (or perhaps, dirty trick) in the past, I'd allowed the "message" portion
to be `None` as a hint that this was going to be a unit-reduction rule. However, I no longer think
that's best. Such renaming-rules are usually optimized out of the tables to begin with. Bracketing
rules are the next most common: those with only one non-void right-hand symbol. They are also a
superset of the few remaining unit-reductions, and common enough in practice.

Anyway, the new idea keeps the two parts of the message separate: `constructor_id, view`.
Bracketing rules are encoded with a negative `constructor_id`, and the combiner is responsible
for positive `constructor_id` numbers.
"""

from ..support import interfaces


def trial_parse(table: interfaces.ParseTable, sentence, *, language=None):
	"""
	This quick-and-dirty trial parser will tell you if a sentence is a member of the language by throwing
	an exception otherwise. It leaves out everything to do with semantic values or parse trees.
	If you want to wrap your head around shift/reduce parsing, this is one way to start.
	"""
	
	def prepare_to_shift(terminal_id) -> int:
		while True:
			step = table.get_action(stack[-1], terminal_id)
			if step > 0: return step  # Shift Action
			elif step == 0:  # Error Action
				stack_symbols = [table.get_breadcrumb(q) for q in stack[1:]]
				raise ValueError(stack_symbols, symbol if terminal_id else interfaces.END_OF_TOKENS, None)
			else:  # Reduce Action
				nonterminal_id, length, constructor_id, view = table.get_rule(-1 - step)
				if length: # Python hiccup: don't let epsilon rules delete the whole stack.
					del stack[ - length:]
				stack.append(table.get_goto(stack[-1], nonterminal_id))
	
	stack = [table.get_initial(language)]
	for symbol in sentence: stack.append(prepare_to_shift(table.get_translation(symbol)))
	prepare_to_shift(table.get_translation(interfaces.END_OF_TOKENS))
	assert len(stack) == 2

class PushDownState:
	"""
	A real treatment of LR-family parsing algorithms should be given in terms
	of that model of computation called a "Push-Down Automaton" (PDA).
	
	In concept, a PDA is a sort of state machine with a stack.
	It has two essential moves:
	* "shift" adds a symbol (with semantic value) to the top of the stack.
	* "reduce" groups sub-phrases into non-terminal symbols at top-of-stack.
	
	In the trial-parse function above, we simulate the PDA with a simple
	list of states. However, for doing any sort of semantic analysis, we're
	going to want something a little nicer: the stack contains both states
	and semantic values under construction.
	
	That's the usual explanation, anyway. In fact "reduce" is better seen as
	a three-part operation: Combine the semantics near the top of the stack,
	pop symbols corresponding to the right-hand side of a production rule,
	and shift the head non-terminal for said rule. Most of that activity is
	more intimately connected with the specific application rather than the
	push-down mechanism, so this class just exposes appropriate primitives.
	The actual "automaton" per-se is the `parse(...)` function, below.
	
	Oh-by-the-way, we don't actually shift symbols per-se, because then
	identifying the correct time to reduce would amount to a search problem.
	Instead, we shift STATES from a characteristic finite-state automaton:
	these encode sufficient left-context to identify the correct places to
	perform a reduction. That FSA forms part of the parse table.
	"""
	
	def __init__(self, initial_state):
		self.stack, self.state = [], initial_state
	
	def shift(self, state, semantic):
		"""
		The most fundamental operation on a PDA is to shift a symbol and a
		semantic value. Actually we don't shift symbols but states; these
		states encode some amount of left-context and are enough to decide
		how to respond to the input.
		"""
		self.stack.append((self.state, semantic))
		self.state = state
	
	def pop_phrase(self, length:int):
		"""
		When a phrase (the right-hand side of a production rule) has been
		recognized, the machine must remove the components of that phrase
		from the stack before shifting the corresponding non-terminal
		symbol (or rather, state) in its place.
		"""
		# Python hiccup: don't let epsilon rules delete the whole stack.
		if length:
			self.state = self.stack[-length][0]
			del self.stack[-length:]
	
	def semantic_view(self, view):
		"""
		Our grammar rules have annotation for which sub-phrases are
		significant to syntax-directed transduction. (Additionally,
		mid-rule actions are shown their significant bits of left-context.)
		The annotations form part of the rule set in the parse tables.
		"""
		return [self.stack[offset][1] for offset in view]
	
	def one_semantic(self, offset:int):
		"""
		There's a common use-case where we just grab the semantic
		value at a specific offset from top-of-stack.
		"""
		return self.stack[offset][1]
	
	def succeed(self):
		""" Fetch the final summary semantic-value. """
		assert len(self.stack) == 1, (self.stack, self.state)
		return self.stack.pop()[1]
	
	def index_state(self, depth:int):
		""" Turns out sometimes we need the state N steps deep in the stack, for error recovery. """
		if depth<=0: return self.state
		else: return self.stack[-depth][0]
	
	def __len__(self):
		""" This is how many steps "deep" you can still find a state. """
		return 1 + len(self.stack)
	
	def path_from_root(self):
		""" Yield each state ID; sometimes useful in error reporting. """
		for state, semantic in self.stack: yield state
		yield self.state
	
class Hypothetical:
	"""
	In the context of error recovery, it's extremely handy to have a general
	"hypothetical parse" facility. This object latches on to an existing PDS
	and acts like a hypothetical branch of that PDS strictly for testing hypothesis.
	"""
	def __init__(self, table:interfaces.ParseTable, pds:PushDownState, initial_depth):
		self.table = table
		self.host = pds
		self.initial_depth = self.watermark = initial_depth
		self.suffix = []
	def tos(self):
		try: return self.suffix[-1]
		except IndexError: return self.host.index_state(self.watermark)
	def shift(self, state): self.suffix.append(state)
	def pop_phrase(self, length:int):
		if length > len(self.suffix):
			self.watermark = self.watermark + length - len(self.suffix)
			self.suffix = []
		elif length: del self.suffix[-length:]
	def reduce(self, rule_id):
		nonterminal_id, length, constructor_id, view = self.table.get_rule(rule_id)
		self.pop_phrase(length)
		self.shift(self.table.get_goto(self.tos(), nonterminal_id))
	def consume(self, terminal_id) -> int:
		""" Returns resulting top-of-stack or raises ValueError. """
		while True:
			step = self.table.get_action(self.tos(), terminal_id)
			if step < 0: self.reduce(-step-1)
			elif step == 0: raise ValueError
			else:
				self.shift(step)
				return step


def parse(table: interfaces.ParseTable, combine, token_stream, *, language=None, on_error:interfaces.ParseErrorListener):
	"""
		:param table: Satisfy the ParseTable interface however you like.
			By reference to the "look-ahead" (terminal) symbol and the current
			state of the PDA, the parse table tells the machine how to act.
		
		:param combine: Gets called with a rule-ID and selected right-hand-side
			semantic elements; must return the semantic value for the left-hand
			side of the corresponding rule. Both mini-parse and the runtime
			support module for MacroParse provide code to help with this part.
		
		:param token_stream: Iterable source of <terminal, semantic> pairs.
			Most normally you'll supply a Scanner, but this version of the
			algorithm makes no such assumption.
		
		:param language: Choice of start-symbol, for multi-language tables.
		
		:return:
			if parsing succeeds: the correct semantic value of the input.
			if error recovery succeeds: the error-ridden semantic value.
			if the error channel propagates an exception: exceptionally.
			otherwise: by `raise interfaces.ParseError(...)`
		
		:param on_error: Once I get error recovery implemented, this
			will be how you direct the parser to report error events.
	"""
	
	def basic_machine_cycle():
		"""
		This is the easy part they explain in all the books...
		
		Classically we might do this with a loop-and-a-half around a stream
		of tokens, treating end-of-text as a sentinel value. But iterators
		are much more the way of Python, so I've twisted the logic up to
		allow the StopIteration exception to do the right thing.
		"""
		for symbol, semantic in token_iterator:
			token_id = table.get_translation(symbol)
			action = find_shift(token_id)
			if action:
				pds.shift(action, semantic)
				perform_immediate_reductions()
			else:
				on_error.unexpected_token(symbol, semantic, pds)
				if not handle_error(token_id, semantic): return
		# After the last real symbol, the parser needs to prepare as if
		# about to shift a notional "end-of-text" symbol -- but don't
		# actually perform that shift: instead that's the signal of
		# an accepted sentence in the language.
		if not find_shift(sentinel_end):
			on_error.unexpected_eof(pds)
			if not handle_error(sentinel_end, None): return
		return pds.succeed()
	
	def find_shift(terminal_id):
		"""
		Before shifting each look-ahead terminal, the parser will reduce
		zero-or-more times until the ACTION table says which state to
		shift into for this terminal.
		"""
		while True:
			action = table.get_action(pds.state, terminal_id)
			if action < 0: reduce_by_rule(-action-1)
			else: return action
	
	def reduce_by_rule(rule_id):
		"""
		When the parser recognizes the right-hand-side of a production rule,
		this is the complete process for converting the sub-phrase to its
		non-terminal symbol on the stack.
		"""
		nonterminal_id, length, constructor_id, view = table.get_rule(rule_id)
		# In a nod to practical brevity, this supports a concise mechanism for
		# (not-eliminated) "unit/renaming" rules and "bracketing" rules:
		#    a negative constructor ID is where in the stack to look for a
		#    representative semantic value.
		if constructor_id < 0: semantic = pds.one_semantic(constructor_id)
		#    otherwise, the constructor ID indexes into a table of messages
		#    which the combiner is responsible to consult.
		else:
			args = pds.semantic_view(view)
			try: semantic = combine(constructor_id, args)
			except Exception as e:
				message = table.get_constructor(constructor_id)
				semantic = on_error.rule_exception(e, message, args)
		pds.pop_phrase(length)
		pds.shift(table.get_goto(pds.state, nonterminal_id), semantic)

	def perform_immediate_reductions():
		"""
		Textbook exposition of shift-reduce parsing may fail to mention the
		practical importance of this step for real-world applications.
		
		Each time the parser shifts a token, it ought to perform interactive
		reductions until another token is strictly necessary to make a
		decision. If you forgo this, then ALL semantic actions (not just the
		ones that hinge on look-ahead) will be delayed by one lexical token.
		This can ruin interactive behavior (think of the desk-calculator)
		and also make it more complicated to report the causes of errors.
		
		It used to be said that this step might impact parser performance.
		Even if true, the difference would be negligible in most contexts.
		"""
		while True:
			action = table.interactive_step(pds.state)
			if action < 0: reduce_by_rule(-action-1)
			else: break
	
	def handle_error(terminal_id, semantic) -> bool:
		"""
		It's time to attempt a parser that can recover smartly from erroneous inputs.
		For a full exposition of the intended strategy, please see the document at:
		https://github.com/kjosib/booze-tools/blob/master/docs/Context%20Free%20Error%20Recovery.md
		
		:return True if the parser is able to resynchronize.
		"""
		# Begin by making a map of the recoverable states.
		#  nb: This relies on dictionary insertion order.
		avenues = {}
		for depth in range(len(pds)): # There's a madness to this un-pythonic method...
			if not table.get_action(pds.index_state(depth), error_token_id): continue
			try: recovery_state = contemplate_recovery(depth, ())
			except ValueError: continue
			if recovery_state not in avenues: avenues[recovery_state] = depth
		if not avenues:
			on_error.cannot_recover()
			return False
		
		# Try each possible re-start against all avenues for recovery.
		# It's a bit brutish, but it works.
		for proposal in generate_proposals(terminal_id, semantic, 3):
			ts, vs = zip(*proposal)
			for recovery_state, depth in avenues.items():
				if table.get_action(recovery_state, ts[0]): # i.e. the proposal has some hope of working here...
					try: contemplate_recovery(depth, ts)
					except ValueError: continue
					else:
						commit_recovery(depth, proposal)
						return True
		on_error.did_not_recover()
		return False
		
	def contemplate_recovery(depth, terminal_ids) -> int:
		h = Hypothetical(table, pds, depth)
		h.consume(error_token_id)
		for t in terminal_ids: h.consume(t)
		return h.tos()
	
	def commit_recovery(depth, proposal):
		err_val = on_error.will_recover(proposal)
		pds.pop_phrase(depth)
		pds.shift(find_shift(error_token_id), err_val)
		for token_id, semantic in proposal:
			pds.shift(find_shift(token_id), semantic)
		if token_id == sentinel_end: pds.pop_phrase(1)
		
	
	def generate_proposals(terminal_id, semantic, length):
		# This generates overlapping rolling runs of tokens.
		proposal = [(terminal_id, semantic)]
		for symbol, value in token_iterator:
			proposal.append((table.get_translation(symbol), value))
			if len(proposal) >= length:
				yield proposal
				proposal.pop(0)
		proposal.append((sentinel_end, None))
		while proposal:
			yield proposal
			proposal.pop(0)
		
	
	sentinel_end = table.get_translation(interfaces.END_OF_TOKENS)
	error_token_id = table.get_translation(interfaces.ERROR_SYMBOL)
	pds = PushDownState(table.get_initial(language))
	token_iterator = iter(token_stream)
	return basic_machine_cycle()





"""
This file aggregates various abstract classes and exception types which BoozeTools deals in.

There's a principle of object-oriented design which says "ask not for data, but for help."
At first glance the ADTs for FiniteAutomaton and ParseTable appear to respect that dictum
only by its violation, as suggested by all these `get_foobar` methods. What gives?

Quite a bit, actually: The scanning and parsing algorithms are data-driven, but the essential
nature of those algorithms should not care about the internal structure and organization of that
data, so long as the proper relevant questions may be answered. This provides the flexibility
to plug in different types of compaction (or no compaction at all) without a complete re-write.

A good modular interface exposes abstract data types and the operations among those types.
The methods on FiniteAutomaton and ParseTable are exactly those needed for the interesting
data-driven algorithms they support, without regard to their internal structure.

On a separate note, you could make a good case for splitting this file in twain. Maybe later.
"""

from . import pretty

END_OF_TOKENS = '<END>' # An agreed artificial "end-of-text" terminal-symbol.
ERROR_SYMBOL = '$error$' # An agreed "error" symbol.
# Note that the scanner should NEVER emit either of the above two symbols.
# However, the error symbol may appear in the right-hand side of a production rule.

DEFAULT_INITIAL_CONDITION = 'INITIAL' # This really is another design constant.

class LanguageError(ValueError):
	""" Base class of all exceptions arising from the language machinery. """

class ScannerBlocked(LanguageError):
	"""
	Raised (by default) if a scanner gets blocked.
	Parameters are:
		the string offset where it happened.
		the current start-condition of the scanner.
	"""
	def __init__(self, position, condition):
		super().__init__(position, condition)
		self.position, self.condition = position, condition

class GeneralizedParseError(LanguageError): pass

class ParseErrorListener:
	"""
	Implement this interface to report/respond to parse errors.
	For the moment I'm assuming you have a handle to the scanner so you
	can get the input-file location of error events...
	"""
	def unexpected_token(self, kind, semantic, pds):
		"""
		The parser has just been given a bogus token.
		It will enter recovery mode next.
		`kind` and `semantic` are whatever the scanner provided.
		`pds` is the state of the push-down automaton at the point of error.
		"""
	
	def unexpected_eof(self, pds):
		"""
		The parser ran out of tokens unexpectedly.
		`pds` is the state of the push-down automaton at the point of error.
		"""
	
	def will_recover(self, tokens):
		"""
		The parser has seen a token sequence sufficient to resynchronize.
		`tokens` is that sequence. The parser will next commit to this
		recovery. (Perhaps there should be a way to prevent it?)
		The return value from this method will appear as the semantic content
		of the "error" position in the error rule that was ultimately chosen.
		"""
	
	def did_not_recover(self):
		"""
		The parser ran out of tokens while in error-recovery mode, and was
		unable to recover.
		"""
	
	def cannot_recover(self):
		"""
		The parser attempted to enter recovery mode, but there are no
		recoverable states on the parse stack, so recovery is impossible.
		
		Default behavior is
		"""
		return self.did_not_recover()
	
	def rule_exception(self, ex:Exception, message, args):
		"""
		Q: If a combining function raises an exception, what should happen?
		A: It depends.
		
		Maybe the exception should not happen: some extra context might help
		you reproduce and debug the problem. Log the context and re-raise.
		
		Maybe certain exceptions represent non-fatal conditions, but you'd
		rather separate policy from mechanism. Deal with it and return the
		semantic value that should replace the aborted attribute-synthesis.
		"""
		raise ex from None # Hide the catch-and-rethrow from the traceback.

class Classifier:
	"""
	Normally a finite-state automaton (FA) based scanner does not treat all possible input
	characters as individual and distinct. Rather, all possible characters are mapped
	to a much smaller alphabet of symbols which are distinguishable from their neighbors
	in terms of their effect on the operation of the FA.

	It is this object's responsibility to perform that mapping via method `classify`.
	"""
	def classify(self, codepoint:int) -> int:
		"""
		Map a unicode codepoint to a specific numbered character class
		such that 0 <= result < self.cardinality()
		as known to a corresponding finite automaton.
		"""
		raise NotImplementedError(type(self))
	def cardinality(self) -> int:
		""" Return the number of distinct classes which may be emitted by self.classify(...). """
		raise NotImplementedError(type(self))
	def display(self):
		""" Pretty-print a suitable representation of the innards of this classifier's data. """
		raise NotImplementedError(type(self))

class FiniteAutomaton:
	"""
	A finite automaton determines which rule matches but knows nothing about the rules themselves.
	This interface captures the operations required to execute the general scanning algorithm.
	It is deliberately decoupled from any particular representation of the underlying data.
	"""
	def jam_state(self): raise NotImplementedError(type(self)) # DFA might provide -1, while NFA might provide an empty frozenset().

	def get_condition(self, condition_name) -> tuple:
		""" A "condition" is implemented as a pair of state_ids for the normal and begining-of-line cases. """
		raise NotImplementedError(type(self))
	
	def get_next_state(self, current_state: int, codepoint: int) -> int:
		""" Does what it says on the tin. codepoint will be -1 at end-of-file, so be prepared. """
		raise NotImplementedError(type(self))
	
	def get_state_rule_id(self, state_id: int) -> int:
		""" Return the associated rule ID if this state is terminal, otherwise None. """
		raise NotImplementedError(type(self))


class ParseTable:
	"""
	This interface captures the operations needed to perform table-driven parsing, as well as a modicum
	of reasonable error reporting. Again, no particular structure or organization is implied.
	"""
	def get_translation(self, symbol) -> int: raise NotImplementedError(type(self, 'Because scanners should not care the order of terminals in the parse table. Zero is reserved for end-of-text.'))
	def get_action(self, state_id:int, terminal_id) -> int: raise NotImplementedError(type(self), 'Positive -> successor state id. Negative -> rule id for reduction. Zero -> error.')
	def get_goto(self, state_id:int, nonterminal_id) -> int: raise NotImplementedError(type(self, 'return a successor state id.'))
	def get_rule(self, rule_id:int) -> tuple: raise NotImplementedError(type(self), 'return a (nonterminal_id, length, constructor_id, view) quad.')
	def get_constructor(self, constructor_id) -> object: raise NotImplementedError(type(self), 'return whatever will make sense to the corresponding combiner.')
	def get_initial(self, language) -> int: raise NotImplementedError(type(self), 'return the initial state id for the selected language, which by the way is usually `None `.')
	def get_breadcrumb(self, state_id:int) -> str: raise NotImplementedError(type(self), 'This is used in error reporting. Return the name of the symbol that shifts into this state.')
	def interactive_step(self, state_id:int) -> int: raise NotImplementedError(type(self), 'Return the reduce instruction for interactive-reducing states; zero otherwise.')
	# These next two methods are in support of GLR parsing:
	def get_split_offset(self) -> int: raise NotImplementedError(type(self), "Action entries >= this number mean to split the parser.")
	def get_split(self, split_id:int) -> list: raise NotImplementedError(type(self), "A list of parse actions of the usual (deterministic) form.")
	
class Scanner:
	"""
	This is the interface a scanner action can expect to be able to operate on.
	
	As a convenience, scan-context stack operations are provided here. There is no "reject" action,
	but a powerful and fast alternative is built into the DFA generator in the form of rule priority
	ranks. The longest-match heuristic breaks ties among the highest ranked rules that match.
	"""
	
	def token(self, kind, semantic=None):
		""" Inform the system that a token of whatever kind and semantic is recognized from the current focus. """
		raise NotImplementedError(type(self))
	
	def enter(self, condition):
		""" Enter the scan condition named by parameter `condition`. """
		raise NotImplementedError(type(self))
	def push(self, condition):
		""" Save the current scan condition to a stack, and enter the scan state named by parameter `condition`. """
		raise NotImplementedError(type(self))
	def pop(self):
		""" Enter the scan condition popped from the top of the stack. """
		raise NotImplementedError(type(self))
	def matched_text(self) -> str:
		""" Return the text currently matched. """
		raise NotImplementedError(type(self))
	def less(self, nr_chars:int):
		""" Put back characters into the stream to be matched: This also provides the mechanism for fixed trailing context. """
		raise NotImplementedError(type(self))
	def current_position(self) -> int:
		""" As advertised. This was motivated by a desire to produce helpful error messages. """
		raise NotImplementedError(type(self))
	def current_span(self):
		""" Return the position and length of the current match-text for use in error-reporting calls and the like. """
		raise NotImplementedError(type(self))
	def current_condition(self) -> str:
		""" Return the most recently entered (or pushed, or popped) start-condition name, which is super-helpful debugging scanners. """
		raise NotImplementedError(type(self))

class ScanRules:
	"""
	The interface a scan-in-progress needs about:
		1. trailing context, and
		2. how to invoke scan rules.
	"""
	
	def get_trailing_context(self, rule_id: int):
		"""
		Fixed trailing context is supported: return a negative number of characters to chop
		from the end of the match. Variable trailing context is supported if the leading stem
		has fixed length: return a positive (or zero) number. The absence of trailing context
		should be indicated by returning None.
		
		It's anticipated that a zero-width stem with trailing context might be used to do things
		like decide which scan state to enter and then restart the scanner from the same point.
		
		It would be a reasonable alternative architecture to design the trailing-context
		support as a separate object in a chain-of-responsibility between the scanner and
		driver. The downside is semantic: trailing context is a feature of the scanner-generator
		and as such should be inseparable from the runtime. It might be a cool performance hack
		to leave out trailing-context support when the feature is not used but that's the sort
		of thing that you build into a tool that generates (e.g.) C code from an automaton.
		
		At some later date, I may decide to add support for both variable leading and trailing
		parts of the same rule. That would mean several additional decisions and added complexity,
		which would definitely justify the alternative design mentioned above.
		"""
		raise NotImplementedError(type(self), ScanRules.get_trailing_context.__doc__)
	
	def invoke(self, yy:Scanner, rule_id:int) -> object:
		"""
		Override this according to your application.
		For example, if you want to emit tokens, call yy.token(kind, semantic)
		"""
		raise NotImplementedError(type(self), ScanRules.invoke.__doc__)

class ScanErrorListener:
	"""
	Implement this interface to report/respond to scan errors.
	For the moment I'm assuming you have a handle to the scanner so you
	can get the input-file location of error events...
	"""
	
	def scan_blocked(self, yy:Scanner):
		"""
		The scanner will call this to report blockage. It will have prepared
		to skip the offending character. Your job is to report the error to
		the user. Try to recover. Emit a designated "nonsense" token and let
		the parser handle it. Delegate to a driver. Do whatever.
		
		Default behavior is to raise an exception, which by the way will kill
		off a parse(...) in progress -- at least until I get parse error
		recovery mode finished.
		"""
		raise ScannerBlocked(yy.current_position(), yy.current_condition())
	
	def scan_exception(self, yy:Scanner, rule_id:int, ex:Exception):
		"""
		If the implementation of scan rule raises an exception, the scanner
		engine will pass that exception to this method (along with its own
		state and the ID number of the failing rule). You're welcome to add
		any sort of context cues, logging, even trying to recover.
		
		If this returns normally, then scanning will resume normally.
		"""
		raise ex from None # Hide the catch-and-rethrow from the traceback.




"""
Even for the most basic unicode support, it's nonsense to work with uncompressed character sets.

As a first (effective, but unicode-naive) approach to working with sets of characters in compact
form, I've chosen to define the character-class data structure as a sorted list of lower bounds
with implied exclusion below the first listed bound. That is: a character is a member of
the class exactly when an odd number of lower-bounds in the class are less-than-or-equal-to
that character's codepoint value. (See the `in_class(...)` function.)

Two major categories of operation exist for these things.

	The first is logical combination of two sets of characters to produce a third.
	The second is to expand classes into a specific set of include/exclude booleans for
	a sorted sequence of bounding points.
	
	The next segment provides for the manufacture of character classes idiomatically:
	singletons, ranges, and logical combinations (set operations).

Of note, I treat -1 as the "end-of-file" character to make end-of-file rule processing
blend in with the rest of the finite-automaton clockwork, but it is a bit special:
"end-of-file" does not appear in the 'universal set' of characters or the complement
of any class, so the only way to get it is expressly from an end-of-file rule or end-of-line
trailing-context. This has implications for how set complements should be constructed.

Lower down in the file, I define standard POSIX-type character classes for the ASCII range.

Note that locale-based POSIX character equivalents are not supported in this module.
Digraphs (e.g. Czech or Spanish "ch") mean the concept works at a higher level than
the individual code point, and would throw several things out of kilter.

It turns out that true unicode character classes (UCCs) have lots and lots and LOTS of
codepoint boundaries, so that it's impractical to use the above idea for supporting
unicode with any sophistication. That calls for a different approach.

Ideally most of the Finite Automaton machinery would be insulated from changes in support
of Smart Unicode Mode. There are certain -- challenges.

"""
import bisect, operator

# How to tell if a character (by codepoint) is a member of the class:
def in_class(cls:list, codepoint:int) -> bool: return bisect.bisect_right(cls, codepoint) % 2


# Character class construction and set-operations:
EMPTY = []
UNIVERSAL = [0]
EOF = [-1, 0]

def singleton(codepoint:int) -> list: return [codepoint, codepoint + 1]
def range_class(first, last) -> list: return [first, last+1] if first <= last else [last, first+1]
def complement(cls:list) -> list:
	if not cls: return UNIVERSAL
	if cls[0]<=0: return cls[1:]
	return [0]+cls
def expand(cls:list, bounds:list):
	""" Given a class and a sorted sequence of codepoints, yield a stream of booleans indicating whether each codepoint is in the class. """
	# It might be simpler to just call 'in_class(...)' in a loop. That would be O(N*log(M)); this is O(N+M).
	# Realistically, the sizes of these things will be small enough to where the simpler approach will be better.
	idx = 0
	for x in bounds:
		while idx<len(cls) and x >= cls[idx]: idx += 1
		yield idx%2
def combine(op, x:list, y:list) -> list:
	""" Arbitrary boolean combination of character classes controlled by 'op :: (bool, bool) -> bool'  """
	result = []
	for b in sorted({0}.union(x, y)): # The zero is included in case op(False, False) == True.
		if len(result) % 2 != bool(op(in_class(x, b), in_class(y, b))):
			result.append(b)
	return result
def union(a:list, b:list) -> list: return combine(operator.or_, a, b)
def intersect(a:list, b:list) -> list: return combine(operator.and_, a, b)
def subtract(a:list, b:list) -> list: return intersect(a, complement(b))

# POSIX classes for the ASCII range:
# (See https://www.regular-expressions.info/posixbrackets.html)

POSIX = {}
POSIX['ascii'] = range_class(0, 127)
POSIX['cntrl'] = union(range_class(0, 31), singleton(127))
POSIX['blank'] = union(singleton(9), singleton(32))
POSIX['space'] = union(range_class(9, 13), singleton(32))
POSIX['digit'] = range_class(ord('0'), ord('9'))
POSIX['upper'] = range_class(ord('A'), ord('Z'))
POSIX['lower'] = range_class(ord('a'), ord('z'))
POSIX['alpha'] = union(POSIX['upper'],  POSIX['lower'])
POSIX['alnum'] = union(POSIX['digit'], POSIX['alpha'])
POSIX['word']  = union(POSIX['alnum'], singleton(ord('_')))
POSIX['xdigit'] = union(POSIX['digit'], union(range_class(ord('A'), ord('F')), range_class(ord('a'), ord('f'))))
POSIX['print'] = subtract(POSIX['ascii'], POSIX['cntrl'])
POSIX['graph'] = subtract(POSIX['print'], POSIX['space'])
POSIX['punct'] = subtract(POSIX['graph'], POSIX['alnum'])

assert all(cls == sorted(cls) for cls in POSIX.values())



""" No frills. Plenty useful. """

from ..support import interfaces, foundation
from . import automata, context_free, shift_reduce


class MiniParse(interfaces.ParseErrorListener):
	""" Connects BNF production rules directly to Python functions. No frills. Very useful as-is. """
	def __init__(self, *start, method='LALR'):
		self.__grammar = context_free.ContextFreeGrammar()
		self.__grammar.start.extend(start)
		self.__hfa: automata.DragonBookTable = None
		self.__awaiting_action = False
		self.__method = method
	
	def left(self, symbols:list): self.__grammar.assoc(context_free.LEFT, symbols)
	def right(self, symbols:list): self.__grammar.assoc(context_free.RIGHT, symbols)
	def nonassoc(self, symbols:list): self.__grammar.assoc(context_free.NONASSOC, symbols)
	
	@staticmethod
	def __analyze(rhs):
		rhs = rhs.split()
		args = []
		for i,symbol in enumerate(rhs):
			if symbol.startswith('.'):
				rhs[i] = symbol[1:]
				args.append(i)
		if not args: args = tuple(range(len(rhs))) # Pick up everything if nothing is special.
		return tuple(rhs), args
	
	def rule(self, lhs:str, rhs:str, prec_sym=None):
		"""
		Decorates a callable as applying when the given rule is recognized.
		Prefix significant arguments with ., or get the whole right-hand side passed as arguments.
		For renaming and bracketing rules, call as
			foo.rule('expr', '( .expr )')(None)
		For normal rules, call as
			@foo.rule('expr', '.expr + .expr')
			def add(a,b): return a+b
		"""
		assert self.__hfa is None
		if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the prior production rule.')
		self.__awaiting_action = True
		rhs, offsets = MiniParse.__analyze(rhs)
		def decorate(fn=None):
			assert self.__awaiting_action
			self.__awaiting_action = False
			if fn is None:
				if len(rhs) == 1: con,plc = None,0 # Unit/renaming rule
				elif len(offsets) == 1: con,plc = None, offsets[0]  # Bracketing rule
				else: con,plc = _collect_tuple, offsets
			else: con,plc = fn, offsets
			self.__grammar.rule(lhs, rhs, prec_sym, con,plc, None)
		return decorate
	
	def renaming(self, lhs:str, *alternatives):
		"""
		Facilitates those "X => A | B | C | D" type rules you often find in real grammars.
		"""
		if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the prior production rule.')
		for branch in alternatives:
			rhs, offsets = MiniParse.__analyze(branch)
			if len(rhs) == 1: con, plc = None, 0  # Unit/renaming rule
			elif len(offsets) == 1: con, plc = None, offsets[0]  # Bracketing rule
			else: raise AssertionError('%r is not a single-member branch -- although you could prepend the significant member with a dot ( like .this ) to fix it.' % branch)
			self.__grammar.rule(lhs, rhs, None, con, plc, None)
	
	def display(self): self.__grammar.display()
	def get_hfa(self, *, strict=False):
		if self.__hfa is None:
			if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the final production rule.')
			self.__grammar.validate()
			parse_style = automata.DeterministicStyle(strict)
			self.__hfa = automata.tabulate(automata.PARSE_TABLE_METHODS[self.__method](self.__grammar), style=parse_style)
		return self.__hfa
	def parse(self, each_token, *, language=None):
		hfa = self.get_hfa()
		constructors = hfa.constructors
		return shift_reduce.parse(hfa, lambda cid,args:constructors[cid](*args), each_token, language=language, on_error=self)
	
def _collect_tuple(*items): return items
